"""unstructured_grid.py: resqpy UnstructuredGrid class module."""

version = '24th November 2021'

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.crs as rqc
import resqpy.olio.triangulation as tri
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.property as rqp
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns

valid_cell_shapes = ['polyhedral', 'tetrahedral', 'pyramidal', 'prism', 'hexahedral']  #: valid cell shapes


class UnstructuredGrid(BaseResqpy):
    """Class for RESQML Unstructured Grid objects."""

    resqml_type = 'UnstructuredGridRepresentation'

    def __init__(self,
                 parent_model,
                 uuid = None,
                 find_properties = True,
                 geometry_required = True,
                 cache_geometry = False,
                 cell_shape = 'polyhedral',
                 title = None,
                 originator = None,
                 extra_metadata = {}):
        """Create an Unstructured Grid object and optionally populate from xml tree.

        arguments:
           parent_model (model.Model object): the model which this grid is part of
           uuid (uuid.UUID, optional): if present, the new grid object is populated from the RESQML object
           find_properties (boolean, default True): if True and uuid is present, a
              grid property collection is instantiated as an attribute, holding properties for which
              this grid is the supporting representation
           geometry_required (boolean, default True): if True and no geometry node exists in the xml,
              an assertion error is raised; ignored if uuid is None
           cache_geometry (boolean, default False): if True and uuid is present, all the geometry arrays
              are loaded into attributes of the new grid object
           cell_shape (str, optional): one of 'polyhedral', 'tetrahedral', 'pyramidal', 'prism', 'hexahedral';
              ignored if uuid is present
           title (str, optional): citation title for new grid; ignored if uuid is present
           originator (str, optional): name of person creating the grid; defaults to login id;
              ignored if uuid is present
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid;
              ignored if uuid is present

        returns:
           a newly created Unstructured Grid object

        notes:
           if not instantiating from an existing RESQML object, then pass a cell_shape here (if geometry is
           going to be used), then set the cell count and, for geometry, build the points array and other
           arrays before writing to the hdf5 file and creating the xml;
           hinge node faces and subnode topology not yet supported at all;
           setting cache_geometry True is equivalent to calling the cache_all_geometry_arrays() method

        :meta common:
        """

        if cell_shape is not None:
            assert cell_shape in valid_cell_shapes, f'invalid cell shape {cell_shape} for unstructured grid'

        self.cell_count = None  #: the number of cells in the grid
        self.cell_shape = cell_shape  #: the shape of cells within the grid
        self.crs_uuid = None  #: uuid of the coordinate reference system used by the grid's geometry
        self.points_cached = None  #: numpy array of raw points data; loaded on demand
        self.node_count = None  #: number of distinct points used in geometry; None if no geometry present
        self.face_count = None  #: number of distinct faces used in geometry; None if no geometry present
        self.nodes_per_face = None
        self.nodes_per_face_cl = None
        self.faces_per_cell = None
        self.cell_face_is_right_handed = None
        self.faces_per_cell_cl = None
        self.inactive = None  #: numpy boolean array indicating which cells are inactive in flow simulation
        self.all_inactive = None  #: boolean indicating whether all cells are inactive
        self.active_property_uuid = None  #: uuid of property holding active cell boolean array (used to populate inactive)
        self.grid_representation = 'UnstructuredGrid'  #: flavour of grid, 'UnstructuredGrid'; not much used
        self.geometry_root = None  #: xml node at root of geometry sub-tree, if present
        self.property_collection = None  #: collection of properties for which this grid is the supporting representation
        self.crs_is_right_handed = None  #: cached boolean indicating handedness of crs axes
        self.cells_per_face = None  #: numpy int array of shape (face_count, 2) holding cells for faces; -1 is null value
        self.xyz_box_cached = None

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if not self.title:
            self.title = 'ROOT'

        if uuid is not None:
            if geometry_required:
                assert self.geometry_root is not None, 'unstructured grid geometry not present in xml'
            if cache_geometry and self.geometry_root is not None:
                self.cache_all_geometry_arrays()
            if find_properties:
                self.extract_property_collection()

    def _load_from_xml(self):
        # Extract simple attributes from xml and set as attributes in this resqpy object
        grid_root = self.root
        assert grid_root is not None
        self.cell_count = rqet.find_tag_int(grid_root, 'CellCount')
        assert self.cell_count > 0
        self.geometry_root = rqet.find_tag(grid_root, 'Geometry')
        if self.geometry_root is None:
            self.cell_shape = None
        else:
            self.extract_crs_uuid()
            self.cell_shape = rqet.find_tag_text(self.geometry_root, 'CellShape')
            assert self.cell_shape in valid_cell_shapes
            self.node_count = rqet.find_tag_int(self.geometry_root, 'NodeCount')
            assert self.node_count > 3
            self.face_count = rqet.find_tag_int(self.geometry_root, 'FaceCount')
            assert self.face_count > 3
        self.extract_inactive_mask()
        # note: geometry arrays not loaded until demanded; see cache_all_geometry_arrays()

    def set_cell_count(self, n: int):
        """Set the number of cells in the grid.

        arguments:
           n (int): the number of cells in the unstructured grid

        note:
           only call this method when creating a new grid, not when working from an existing RESQML grid
        """
        assert self.cell_count is None or self.cell_count == n
        self.cell_count = n

    def active_cell_count(self):
        """Returns the number of cells deemed to be active for flow simulation purposes."""
        if self.inactive is None:
            return self.cell_count
        return self.cell_count - np.count_nonzero(self.inactive)

    def cache_all_geometry_arrays(self):
        """Loads from hdf5 into memory all the arrays defining the grid geometry.

        returns:
           None

        notes:
           call this method if much grid geometry processing is coming up;
           the arrays are cached as direct attributes to this grid object;
           the node and face indices make use of 'jagged' arrays (effectively an array of lists represented as
           a linear array and a 'cumulative length' index array)

        :meta common:
        """

        assert self.node_count is not None and self.face_count is not None

        self.points_ref()

        if self.nodes_per_face is None:
            self._load_jagged_array('NodesPerFace', 'nodes_per_face')
            assert len(self.nodes_per_face_cl) == self.face_count

        if self.faces_per_cell is None:
            self._load_jagged_array('FacesPerCell', 'faces_per_cell')
            assert len(self.faces_per_cell_cl) == self.cell_count

        if self.cell_face_is_right_handed is None:
            assert self.geometry_root is not None
            cfirh_node = rqet.find_tag(self.geometry_root, 'CellFaceIsRightHanded')
            assert cfirh_node is not None
            h5_key_pair = self.model.h5_uuid_and_path_for_node(cfirh_node)
            self.model.h5_array_element(h5_key_pair,
                                        index = None,
                                        cache_array = True,
                                        object = self,
                                        array_attribute = 'cell_face_is_right_handed',
                                        required_shape = (len(self.faces_per_cell),),
                                        dtype = 'bool')

    def _load_jagged_array(self, tag, main_attribute):
        # jagged arrays are used by RESQML to efficiantly pack arrays of lists of numbers
        assert self.geometry_root is not None
        root_node = rqet.find_tag(self.geometry_root, tag)
        assert root_node is not None
        elements_root = rqet.find_tag(root_node, 'Elements')
        h5_key_pair = self.model.h5_uuid_and_path_for_node(elements_root)
        self.model.h5_array_element(h5_key_pair,
                                    index = None,
                                    cache_array = True,
                                    object = self,
                                    array_attribute = main_attribute,
                                    dtype = 'int')
        cum_length_root = rqet.find_tag(root_node, 'CumulativeLength')
        h5_key_pair = self.model.h5_uuid_and_path_for_node(cum_length_root)
        self.model.h5_array_element(h5_key_pair,
                                    index = None,
                                    cache_array = True,
                                    object = self,
                                    array_attribute = main_attribute + '_cl',
                                    dtype = 'int')

    def extract_crs_uuid(self):
        """Returns uuid for coordinate reference system, as stored in geometry xml tree.

        returns:
           uuid.UUID object
        """

        if self.crs_uuid is not None:
            return self.crs_uuid
        if self.geometry_root is None:
            return None
        uuid_str = rqet.find_nested_tags_text(self.geometry_root, ['LocalCrs', 'UUID'])
        if uuid_str:
            self.crs_uuid = bu.uuid_from_string(uuid_str)
            self._set_crs_handedness()
        return self.crs_uuid

    def extract_inactive_mask(self):
        """Returns boolean numpy array indicating which cells are inactive, if (in)active property found for this grid.

        returns:
           numpy array of booleans, of shape (cell_count,) being True for cells which are inactive; False for active

        note:
           RESQML does not have a built-in concept of inactive (dead) cells, though the usage guide advises to use a
           discrete property with a local property kind of 'active'; this resqpy code can maintain an 'inactive'
           attribute for the grid object, which is a boolean numpy array indicating which cells are inactive
        """

        if self.inactive is not None:
            return self.inactive
        self.inactive = np.zeros((self.cell_count,), dtype = bool)  # ie. all active
        self.all_inactive = False
        gpc = self.extract_property_collection()
        if gpc is None:
            return self.inactive
        active_gpc = rqp.PropertyCollection()
        # note: use of bespoke (local) property kind 'active' as suggested in resqml usage guide
        active_gpc.inherit_parts_selectively_from_other_collection(other = gpc,
                                                                   property_kind = 'active',
                                                                   continuous = False)
        if active_gpc.number_of_parts() > 0:
            if active_gpc.number_of_parts() > 1:
                log.warning('more than one property found with bespoke kind "active", using last encountered')
            active_part = active_gpc.parts()[-1]
            active_array = active_gpc.cached_part_array_ref(active_part, dtype = 'bool')
            self.inactive = np.logical_not(active_array)
            self.active_property_uuid = active_gpc.uuid_for_part(active_part)
            active_gpc.uncache_part_array(active_part)
            self.all_inactive = np.all(self.inactive)

        return self.inactive

    def extract_property_collection(self):
        """Load grid property collection object holding lists of all properties in model that relate to this grid.

        returns:
           resqml_property.PropertyCollection object

        note:
           a reference to the grid property collection is cached in this grid object; if the properties change,
           for example by generating some new properties, the property_collection attribute of the grid object
           would need to be reset to None elsewhere before calling this method again
        """

        if self.property_collection is not None:
            return self.property_collection
        self.property_collection = rqp.PropertyCollection(support = self)
        return self.property_collection

    def set_cells_per_face(self, check_all_faces_used = True):
        """Sets and returns the cells_per_face array showing which cells are using each face.

        arguments:
           check_all_faces_used (boolean, default True): if True, an assertion error is raised if there are any
              faces which do not appear in any cells

        returns:
           numpy int array of shape (face_count, 2) showing upto 2 cell indices for each face index; -1 is a
           null value; if only one cell uses a face, its index is always in position 0 of the second axis
        """

        if self.cells_per_face is not None:
            return self.cells_per_face
        assert self.face_count is not None
        self.cells_per_face = -np.ones((self.face_count, 2), dtype = int)  # -1 is used here as a null value
        for cell in range(self.cell_count):
            for face in self.face_indices_for_cell(cell):
                if self.cells_per_face[face, 0] == -1:
                    self.cells_per_face[face, 0] = cell
                else:
                    assert self.cells_per_face[face, 1] == -1, f'more than two cells use face with index {face}'
                    self.cells_per_face[face, 1] = cell
        if check_all_faces_used:
            assert np.all(self.cells_per_face[:, 0] >= 0), 'not all faces used by cells'
        return self.cells_per_face

    def masked_cells_per_face(self, exclude_cell_mask):
        """Sets and returns the cells_per_face array showing which cells are using each face.

        arguments:
           exclude_cell_mask (numpy bool array of shape (cell_count,)): cells with a value True in this array
           are excluded when populating the result

        returns:
           numpy int array of shape (face_count, 2) showing upto 2 cell indices for each face index; -1 is a
           null value; if only one cell uses a face, its index is always in position 0 of the second axis

        note:
           this method recomputes the result on every call - nothing extra is cached in the grid
        """

        assert self.face_count is not None
        result = -np.ones((self.face_count, 2), dtype = int)  # -1 is used here as a null value
        for cell in range(self.cell_count):
            if exclude_cell_mask[cell]:
                continue
            for face in self.face_indices_for_cell(cell):
                if result[face, 0] == -1:
                    result[face, 0] = cell
                else:
                    assert result[face, 1] == -1, f'more than two cells use face with index {face}'
                    result[face, 1] = cell
        return result

    def external_face_indices(self):
        """Returns a numpy int vector listing the indices of faces which are only used by one cell.

        note:
           resulting array is ordered by face index
        """

        self.set_cells_per_face()
        return np.where(self.cells_per_face[:, 1] == -1)[0]

    def external_face_indices_for_masked_cells(self, exclude_cell_mask):
        """Returns a numpy int vector listing the indices of faces which are used by exactly one of masked cells.

        arguments:
           exclude_cell_mask (numpy bool array of shape (cell_count,)): cells with a value True in this array
           are excluded when populating the result

        note:
           resulting array is ordered by face index
        """

        cpf = self.masked_cells_per_face(exclude_cell_mask)
        return np.where(np.logical_and(cpf[:, 0] >= 0, cpf[:, 1] == -1))[0]

    def internal_face_indices_for_masked_cells(self, exclude_cell_mask):
        """Returns a numpy int vector listing the indices of faces which are used by two of masked cells.

        arguments:
           exclude_cell_mask (numpy bool array of shape (cell_count,)): cells with a value True in this array
           are excluded when populating the result

        note:
           resulting array is ordered by face index
        """

        cpf = self.masked_cells_per_face(exclude_cell_mask)
        return np.where(np.logical_and(cpf[:, 0] >= 0, cpf[:, 1] >= 0))[0]

    def points_ref(self):
        """Returns an in-memory numpy array containing the xyz data for points used in the grid geometry.

        returns:
           numpy array of shape (node_count, 3)

        notes:
           this is the usual way to get at the actual grid geometry points data in the native RESQML layout;
           the array is cached as an attribute of the grid object

        :meta common:
        """

        if self.points_cached is None:

            assert self.node_count is not None

            p_root = rqet.find_tag(self.geometry_root, 'Points')
            if p_root is None:
                log.debug('points_ref() returning None as geometry not present')
                return None  # geometry not present

            assert rqet.node_type(p_root) == 'Point3dHdf5Array'
            h5_key_pair = self.model.h5_uuid_and_path_for_node(p_root, tag = 'Coordinates')
            if h5_key_pair is None:
                return None

            self.model.h5_array_element(h5_key_pair,
                                        index = None,
                                        cache_array = True,
                                        object = self,
                                        array_attribute = 'points_cached',
                                        required_shape = (self.node_count, 3))

        return self.points_cached

    def xyz_box(self):
        """Returns the minimum and maximum xyz for the grid geometry.

        returns:
           numpy array of float of shape (2, 3); the first axis is minimum, maximum; the second axis is x, y, z

        :meta common:
        """

        if self.xyz_box_cached is None:
            self.xyz_box_cached = np.empty((2, 3), dtype = float)
            p = self.points_ref()
            self.xyz_box_cached[0] = np.min(p, axis = 0)
            self.xyz_box_cached[1] = np.max(p, axis = 0)

        return self.xyz_box_cached

    def face_centre_point(self, face_index):
        """Returns a nominal centre point for a single face calculated as the mean position of its nodes.

        arguments:
           face_index (int): the index of the face (as used in faces_per_cell and implicitly in nodes_per_face)

        returns:
           numpy float array of shape (3,) being the xyz location of the centre point of the face

        note:
           this returns a nominal centre point for a face - the mean position of its nodes - which is not generally
           its barycentre
        """

        return np.mean(self.points_cached[self.node_indices_for_face(face_index)], axis = 0)

    def face_count_for_cell(self, cell):
        """Returns the number of faces for a particular cell."""

        self.cache_all_geometry_arrays()
        start = 0 if cell == 0 else self.faces_per_cell_cl[cell - 1]
        return self.faces_per_cell_cl[cell] - start

    def max_face_count_for_any_cell(self):
        """Returns the largest number of faces in use by any one cell."""

        self.cache_all_geometry_arrays()
        return max(self.faces_per_cell_cl[0], np.max(self.faces_per_cell_cl[1:] - self.faces_per_cell_cl[:-1]))

    def node_count_for_face(self, face_index):
        """Returns the number of nodes for a particular face."""

        self.cache_all_geometry_arrays()
        start = 0 if face_index == 0 else self.nodes_per_face_cl[face_index - 1]
        return self.nodes_per_face_cl[face_index] - start

    def max_node_count_for_any_face(self):
        """Returns the largest number of nodes in use by any one face."""

        self.cache_all_geometry_arrays()
        return max(self.nodes_per_face_cl[0], np.max(self.nodes_per_face_cl[1:] - self.nodes_per_face_cl[:-1]))

    def node_indices_for_face(self, face_index):
        """Returns numpy list of node indices for a single face.

        arguments:
           face_index (int): the index of the face (as used in faces_per_cell and implicitly in nodes_per_face)

        returns:
           numpy int array of shape (N,) being the node indices identifying the vertices of the face

        note:
           the node indices are used to index the points data (points_cached attribute);
           ordering of returned nodes is clockwise or anticlockwise when viewed from within the cell,
           as indicated by the entry in the cell_face_is_right_handed array
        """

        self.cache_all_geometry_arrays()
        start = 0 if face_index == 0 else self.nodes_per_face_cl[face_index - 1]
        return self.nodes_per_face[start:self.nodes_per_face_cl[face_index]].copy()

    def distinct_node_indices_for_cell(self, cell):
        """Returns a numpy list of distinct node indices used by the faces of a single cell.

        arguments:
           cell (int): the index of the cell for which distinct node indices are required

        returns:
           numpy int array of shape (N, ) being the indices of N distinct nodes used by the cell's faces

        note:
           the returned array is sorted by increasing node index
        """

        face_indices = self.face_indices_for_cell(cell)
        node_set = self.node_indices_for_face(face_indices[0])
        for face_index in face_indices[1:]:
            node_set = np.union1d(node_set, self.node_indices_for_face(face_index))
        return node_set

    def face_indices_for_cell(self, cell):
        """Returns numpy list of face indices for a single cell.

        arguments:
           cell (int): the index of the cell for which face indices are required

        returns:
           numpy int array of shape (F,) being the face indices of each of the F faces for the cell

        note:
           the face indices are used when accessing the nodes per face data and can also be used to identify
           shared faces
        """

        self.cache_all_geometry_arrays()
        start = 0 if cell == 0 else self.faces_per_cell_cl[cell - 1]
        return self.faces_per_cell[start:self.faces_per_cell_cl[cell]].copy()

    def face_indices_and_handedness_for_cell(self, cell):
        """Returns numpy list of face indices for a single cell, and numpy boolean list of face right handedness.

        arguments:
           cell (int): the index of the cell for which face indices are required

        returns:
           numpy int array of shape (F,), numpy boolean array of shape (F, ):
           being the face indices of each of the F faces for the cell, and the right handedness (clockwise order)
           of the face nodes when viewed from within the cell

        note:
           the face indices are used when accessing the nodes per face data and can also be used to identify
           shared faces; the handedness (clockwise or anti-clockwise ordering of nodes) is significant for
           some processing of geometry such as volume calculations
        """

        self.cache_all_geometry_arrays()
        start = 0 if cell == 0 else self.faces_per_cell_cl[cell - 1]
        return (self.faces_per_cell[start:self.faces_per_cell_cl[cell]].copy(),
                self.cell_face_is_right_handed[start:self.faces_per_cell_cl[cell]].copy())

    def edges_for_face(self, face_index):
        """Returns numpy list of pairs of node indices, each pair being one edge of the face.

        arguments:
           face_index (int): the index of the face (as used in faces_per_cell and implicitly in nodes_per_face)

        returns:
           numpy int array of shape (N, 2) being the node indices identifying the N edges of the face

        notes:
           the order of the pairs follows the order of the nodes for the face; within each pair, the
           order of the two node indices also follows the order of the nodes for the face
        """

        face_nodes = self.node_indices_for_face(face_index)
        return np.array([(face_nodes[i - 1], face_nodes[i]) for i in range(len(face_nodes))], dtype = int)

    def edges_for_face_with_node_indices_ordered_within_pairs(self, face_index):
        """Returns numpy list of pairs of node indices, each pair being one edge of the face.

        arguments:
           face_index (int): the index of the face (as used in faces_per_cell and implicitly in nodes_per_face)

        returns:
           numpy int array of shape (N, 2) being the node indices identifying the N edges of the face

        notes:
           the order of the pairs follows the order of the nodes for the face; within each pair, the
           two node indices are ordered with the lower index first
        """

        edges = self.edges_for_face(face_index)
        for i in range(len(edges)):
            a, b = edges[i]
            if b < a:
                edges[i] = (b, a)
        return edges

    def distinct_edges_for_cell(self, cell):
        """Returns numpy list of pairs of node indices, each pair being one distinct edge of the cell.

        arguments:
           cell (int): the index of the cell

        returns:
           numpy int array of shape (E, 2) being the node indices identifying the E edges of the cell

        note:
           within each pair, the two node indices are ordered with the lower index first
        """

        edge_list = []
        for face_index in self.face_indices_for_cell(cell):
            for a, b in self.edges_for_face(face_index):
                if b < a:
                    a, b = b, a
                if (a, b) not in edge_list:
                    edge_list.append((a, b))
        return np.array(edge_list, dtype = int)

    def cell_face_centre_points(self, cell):
        """Returns a numpy array of centre points of the faces for a single cell.

        arguments:
           cell (int): the index of the cell for which face centre points are required

        returns:
           numpy array of shape (F, 3) being the xyz location of each of the F faces for the cell

        notes:
           the order of the returned face centre points matches the faces_per_cell for the cell;
           the returned values are nominal centre points for the faces - the mean position of their nodes - which
           are not generally their barycentres
        """

        face_indices = self.face_indices_for_cell(cell)
        face_centres = np.empty((len(face_indices), 3))
        for fi, face_index in enumerate(face_indices):  # todo: vectorise?
            face_centres[fi] = self.face_centre_point(face_index)
        return face_centres

    def face_normal(self, face_index):
        """Returns a unit vector normal to a planar approximation of the face.

        arguments:
           face_index (int): the index of the face (as used in faces_per_cell and implicitly in nodes_per_face)

        returns:
           numpy float array of shape (3,) being the xyz components of a unit length vector normal to the face

        note:
           in the case of a degenerate face, a zero length vector is returned;
           the direction of the normal will be into or out of the cell depending on the handedness of the
           cell face
        """

        self.cache_all_geometry_arrays()
        vertices = self.points_cached[self.node_indices_for_face(face_index)]
        centre = self.face_centre_point(face_index)
        normal_sum = np.zeros(3)

        for e in range(len(vertices)):
            edge = vertices[e] - vertices[e - 1]
            radial = centre - 0.5 * (vertices[e] + vertices[e - 1])
            weight = vec.naive_length(edge)
            if weight == 0.0:
                continue
            edge_normal = vec.unit_vector(vec.cross_product(edge, radial))
            normal_sum += weight * edge_normal

        return vec.unit_vector(normal_sum)

    def planar_face_points(self, face_index, xy_plane = False):
        """Returns points for a planar approximation of a face.

        arguments:
           face_index (int): the index of the face for which a planar approximation is required
           xy_plane (boolean, default False): if True, the returned points lie in a horizontal plane with z = 0.0;
              if False, the plane is located approximately in the position of the original face, with the same
              normal direction

        returns:
           numpy float array of shape (N, 3) being the xyz points of the planar face nodes corresponding to the
           N nodes of the original face, in the same order
        """

        self.cache_all_geometry_arrays()
        face_centre = self.face_centre_point(face_index)
        assert np.all(self.node_indices_for_face(face_index) >= 0)  # debug
        assert np.all(self.node_indices_for_face(face_index) < self.node_count)  # debug
        face_points = self.points_cached[self.node_indices_for_face(face_index), :].copy()
        normal = self.face_normal(face_index)
        az = vec.azimuth(normal)
        incl = vec.inclination(normal)
        vec.tilt_points(face_centre, az, -incl, face_points)  # modifies face_points in situ
        if xy_plane:
            face_points[..., 2] = 0.0
        else:
            face_points[..., 2] = face_centre[2]
            vec.tilt_points(face_centre, az, incl, face_points)  # tilt back to original average normal
        return face_points

    def face_triangulation(self, face_index, local_nodes = False):
        """Returns a Delauney triangulation of (a planar approximation of) a face.

        arguments:
           face_index (int): the index of the face for which a triangulation is required
           local_nodes (boolean, default False): if True, the returned node indices are local to the face nodes,
              ie. can index into node_indices_for_face(); if False, the returned node indices are the global
              node indices in use by the grid

        returns:
           numpy int array of shape (N - 2, 3) being the node indices of the triangulation, where N is the number
           of nodes defining the face
        """

        face_points = self.planar_face_points(face_index, xy_plane = True)
        local_triangulation = tri.dt(face_points)  # returns int array of shape (M, 3)
        assert len(
            local_triangulation) == len(face_points) - 2, 'face triangulation failure (concave edges when planar?)'
        if local_nodes:
            return local_triangulation
        return self.node_indices_for_face(face_index)[local_triangulation]

    def area_of_face(self, face_index, in_plane = False):
        """Returns the area of a face.

        arguments:
           face_index (int): the index of the face for which the area is required
           in_plane (boolean, default False): if True, the area returned is the area of the planar approximation
              of the face; if False, the area is the sum of the areas of the triangulation of the face, which
              need not be planar

        returns:
           float being the area of the face

        notes:
           units of measure of the area is implied by the units of the crs in use by the grid
        """

        if in_plane:
            face_points = self.planar_face_points(face_index, xy_plane = True)
            local_triangulation = tri.dt(face_points)
            triangulated_points = face_points[local_triangulation]
        else:
            global_triangulation = self.face_triangulation(face_index)
            triangulated_points = self.points_cached[global_triangulation]
        assert triangulated_points.ndim == 3 and triangulated_points.shape[1:] == (3, 3)
        area = 0.0
        for tp in triangulated_points:
            area += vec.area_of_triangle(tp[0], tp[1], tp[2])
        return area

    def cell_centre_point(self, cell):
        """Returns centre point of a single cell calculated as the mean position of the centre points of its faces.

        arguments:
           cell (int): the index of the cell for which the centre point is required

        returns:
           numpy float array of shape (3,) being the xyz location of the centre point of the cell

        note:
           this is a nominal centre point - the mean of the nominal face centres - which is not generally
           the barycentre of the cell
        """

        return np.mean(self.cell_face_centre_points(cell), axis = 0)

    def centre_point(self, cell = None, cache_centre_array = False):
        """Returns centre point of a cell or array of centre points of all cells.
        
        Optionally cache centre points for all cells.

        arguments:
           cell (optional): if present, the cell number of the individual cell for which the
              centre point is required; zero based indexing
           cache_centre_array (boolean, default False): If True, or cell is None, an array of centre points
              is generated and added as an attribute of the grid, with attribute name array_centre_point

        returns:
           (x, y, z) 3 element numpy array of floats holding centre point of a single cell;
           or numpy 2D array of shape (cell_count, 3) if cell is None

        notes:
           a simple mean of the nominal centres of the faces is used to calculate the centre point of a cell;
           this is not generally the barycentre of the cell;
           resulting coordinates are in the same (local) crs as the grid points

        :meta common:
        """

        if cell is None:
            cache_centre_array = True

        if hasattr(self, 'array_centre_point') and self.array_centre_point is not None:
            if cell is None:
                return self.array_centre_point
            return self.array_centre_point[cell]  # could check for nan here and return None

        if self.node_count is None:  # no geometry present
            return None

        if cache_centre_array:  # calculate for all cells and cache
            self.array_centre_point = np.empty((self.cell_count, 3))
            for cell_index in range(self.cell_count):  # todo: vectorise
                self.array_centre_point[cell_index] = self.cell_centre_point(cell_index)
            if cell is None:
                return self.array_centre_point
            else:
                return self.array_centre_point[cell]

        else:
            return self.cell_centre_point(cell)

    def volume(self, cell):
        """Returns the volume of a single cell.

        arguments:
           cell (int): the index of the cell for which the volume is required

        returns:
           float being the volume of the cell; units of measure is implied by crs units

        note:
           this is a computationally expensive method
        """

        from resqpy.unstructured._tetra_grid import TetraGrid

        tetra = TetraGrid.from_unstructured_cell(self, cell, set_handedness = False)
        assert tetra is not None
        return tetra.grid_volume()

    def check_indices(self):
        """Asserts that all node and face indices are within range."""

        self.cache_all_geometry_arrays()
        assert self.cell_count > 0
        assert self.face_count >= 4
        assert self.node_count >= 4
        assert np.all(self.faces_per_cell >= 0) and np.all(self.faces_per_cell < self.face_count)
        assert self.faces_per_cell_cl[0] >= 4
        assert np.all(self.faces_per_cell_cl[1:] - self.faces_per_cell_cl[:-1] >= 4)
        assert len(self.faces_per_cell_cl) == self.cell_count
        assert np.all(self.nodes_per_face >= 0) and np.all(self.nodes_per_face < self.node_count)
        assert self.nodes_per_face_cl[0] >= 3
        assert np.all(self.nodes_per_face_cl[1:] - self.nodes_per_face_cl[:-1] >= 3)
        assert len(self.nodes_per_face_cl) == self.face_count

    def xy_units(self):
        """Returns the projected view (x, y) units of measure of the coordinate reference system for the grid.

        :meta common:
        """

        return rqet.find_tag(self._crs_root(), 'ProjectedUom').text

    def z_units(self):
        """Returns the vertical (z) units of measure of the coordinate reference system for the grid.

        :meta common:
        """

        return rqet.find_tag(self._crs_root(), 'VerticalUom').text

    def write_hdf5(self, file = None, geometry = True, imported_properties = None, write_active = None):
        """Write to an hdf5 file the datasets for the grid geometry and optionally properties from cached arrays.

        :meta common:
        """

        # NB: when writing a new geometry, all arrays must be set up and exist as the appropriate attributes prior to calling this function
        # if saving properties, active cell array should be added to imported_properties based on logical negation of inactive attribute
        # xml is not created here for property objects

        if geometry:
            assert self.node_count > 0 and self.face_count > 0, 'geometry not present when writing unstructured grid to hdf5'

        if write_active is None:
            write_active = geometry

        self.cache_all_geometry_arrays()

        if not file:
            file = self.model.h5_file_name()
        h5_reg = rwh5.H5Register(self.model)

        if geometry:
            h5_reg.register_dataset(self.uuid, 'Points', self.points_cached)
            h5_reg.register_dataset(self.uuid, 'NodesPerFace/elements', self.nodes_per_face, dtype = 'uint32')
            h5_reg.register_dataset(self.uuid,
                                    'NodesPerFace/cumulativeLength',
                                    self.nodes_per_face_cl,
                                    dtype = 'uint32')
            h5_reg.register_dataset(self.uuid, 'FacesPerCell/elements', self.faces_per_cell, dtype = 'uint32')
            h5_reg.register_dataset(self.uuid,
                                    'FacesPerCell/cumulativeLength',
                                    self.faces_per_cell_cl,
                                    dtype = 'uint32')
            h5_reg.register_dataset(self.uuid, 'CellFaceIsRightHanded', self.cell_face_is_right_handed, dtype = 'uint8')

        if write_active and self.inactive is not None:
            if imported_properties is None:
                imported_properties = rqp.PropertyCollection()
                imported_properties.set_support(support = self)
            else:
                filtered_list = []
                for entry in imported_properties.imported_list:
                    if entry[2].upper() == 'ACTIVE' or entry[10] == 'active':
                        continue  # keyword or property kind
                    filtered_list.append(entry)
                imported_properties.imported_list = filtered_list  # might have unintended side effects elsewhere
            active_mask = np.logical_not(self.inactive)
            imported_properties.add_cached_array_to_imported_list(active_mask,
                                                                  'active cell mask',
                                                                  'ACTIVE',
                                                                  discrete = True,
                                                                  property_kind = 'active')

        if imported_properties is not None and imported_properties.imported_list is not None:
            for entry in imported_properties.imported_list:
                if hasattr(imported_properties, entry[3]):  # otherwise constant array
                    h5_reg.register_dataset(entry[0], 'values_patch0', imported_properties.__dict__[entry[3]])
                if entry[10] == 'active':
                    self.active_property_uuid = entry[0]

        h5_reg.write(file, mode = 'a')

    def create_xml(self,
                   ext_uuid = None,
                   add_as_part = True,
                   add_relationships = True,
                   title = None,
                   originator = None,
                   write_active = True,
                   write_geometry = True,
                   extra_metadata = {}):
        """Creates an unstructured grid node and optionally adds as a part in the model.

        arguments:
           ext_uuid (uuid.UUID, optional): the uuid of the hdf5 external part holding the array data for the grid geometry
           add_as_part (boolean, default True): if True, the newly created xml node is added as a part
              in the model
           add_relationships (boolean, default True): if True, relationship xml parts are created relating the
              new grid part to: the crs, and the hdf5 external part
           title (string): used as the citation title text; careful consideration should be given
              to this argument when dealing with multiple grids in one model, as it is the means by which a
              human will distinguish them
           originator (string, optional): the name of the human being who created the unstructured grid part;
              default is to use the login name
           write_active (boolean, default True): if True, xml for an active cell property is also generated, but
              only if the active_property_uuid is set and no part exists in the model for that uuid
           write_geometry (boolean, default True): if False, the geometry node is omitted from the xml
           extra_metadata (dict): any key value pairs in this dictionary are added as extra metadata xml nodes

        returns:
           the newly created unstructured grid xml node

        notes:
           the write_active argument should generally be set to the same value as that passed to the write_hdf5... method;
           the RESQML standard allows the geometry to be omitted for a grid, controlled here by the write_geometry argument;
           the explicit geometry may be omitted for unstructured grids, in which case the arrays should not be written to
           the hdf5 file either

        :meta common:
        """

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()
        if title:
            self.title = title
        if not self.title:
            self.title = 'ROOT'

        ug = super().create_xml(add_as_part = False, originator = originator, extra_metadata = extra_metadata)

        if self.grid_representation and not write_geometry:
            rqet.create_metadata_xml(node = ug, extra_metadata = {'grid_flavour': self.grid_representation})

        cc_node = rqet.SubElement(ug, ns['resqml2'] + 'CellCount')
        cc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        cc_node.text = str(self.cell_count)

        if write_geometry:

            geom = rqet.SubElement(ug, ns['resqml2'] + 'Geometry')
            geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'UnstructuredGridGeometry')
            geom.text = '\n'

            # the remainder of this function is populating the geometry node
            self.model.create_crs_reference(crs_uuid = self.crs_uuid, root = geom)

            points_node = rqet.SubElement(geom, ns['resqml2'] + 'Points')
            points_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
            points_node.text = '\n'

            coords = rqet.SubElement(points_node, ns['resqml2'] + 'Coordinates')
            coords.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            coords.text = '\n'

            self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'Points', root = coords)

            shape_node = rqet.SubElement(geom, ns['resqml2'] + 'CellShape')
            shape_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'CellShape')
            shape_node.text = self.cell_shape

            nc_node = rqet.SubElement(geom, ns['resqml2'] + 'NodeCount')
            nc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            nc_node.text = str(self.node_count)

            fc_node = rqet.SubElement(geom, ns['resqml2'] + 'FaceCount')
            fc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
            fc_node.text = str(self.face_count)

            self._create_jagged_array_xml(geom, 'NodesPerFace', ext_uuid)

            self._create_jagged_array_xml(geom, 'FacesPerCell', ext_uuid)

            cfirh_node = rqet.SubElement(geom, ns['resqml2'] + 'CellFaceIsRightHanded')
            cfirh_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'BooleanHdf5Array')
            cfirh_node.text = '\n'

            cfirh_values = rqet.SubElement(cfirh_node, ns['resqml2'] + 'Values')
            cfirh_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            cfirh_values.text = '\n'

            self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'CellFaceIsRightHanded', root = cfirh_values)

            self.geometry_root = geom

        if add_as_part:
            self.model.add_part('obj_UnstructuredGridRepresentation', self.uuid, ug)
            if add_relationships:
                if write_geometry:
                    # create 2 way relationship between UnstructuredGrid and Crs
                    self.model.create_reciprocal_relationship(ug, 'destinationObject',
                                                              self.model.root(uuid = self.crs_uuid), 'sourceObject')
                    # create 2 way relationship between UnstructuredGrid and Ext
                    ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                    ext_node = self.model.root_for_part(ext_part)
                    self.model.create_reciprocal_relationship(ug, 'mlToExternalPartProxy', ext_node,
                                                              'externalPartProxyToMl')

        if write_active and self.active_property_uuid is not None and self.model.part(
                uuid = self.active_property_uuid) is None:
            active_collection = rqp.PropertyCollection()
            active_collection.set_support(support = self)
            active_collection.create_xml(None,
                                         None,
                                         'ACTIVE',
                                         'active',
                                         p_uuid = self.active_property_uuid,
                                         discrete = True,
                                         add_min_max = False,
                                         find_local_property_kinds = True)

        return ug

    def _create_jagged_array_xml(self, parent_node, tag, ext_uuid, null_value = -1):

        j_node = rqet.SubElement(parent_node, ns['resqml2'] + tag)
        j_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ResqmlJaggedArray')
        j_node.text = '\n'

        elements = rqet.SubElement(j_node, ns['resqml2'] + 'Elements')
        elements.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
        elements.text = '\n'

        el_null = rqet.SubElement(elements, ns['resqml2'] + 'NullValue')
        el_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        el_null.text = str(null_value)

        el_values = rqet.SubElement(elements, ns['resqml2'] + 'Values')
        el_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        el_values.text = '\n'

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, tag + '/elements', root = el_values)

        c_length = rqet.SubElement(j_node, ns['resqml2'] + 'CumulativeLength')
        c_length.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
        c_length.text = '\n'

        cl_null = rqet.SubElement(c_length, ns['resqml2'] + 'NullValue')
        cl_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        cl_null.text = '0'

        cl_values = rqet.SubElement(c_length, ns['resqml2'] + 'Values')
        cl_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        cl_values.text = '\n'

        self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, tag + '/cumulativeLength', root = cl_values)

    def _set_crs_handedness(self):
        if self.crs_is_right_handed is not None:
            return
        assert self.crs_uuid is not None
        crs = rqc.Crs(self.model, uuid = self.crs_uuid)
        self.crs_is_right_handed = crs.is_right_handed_xyz()

    def _crs_root(self):
        """Returns xml root node for the crs object referred to by this grid."""

        if self.crs_uuid is None:
            return None
        return self.model.root(uuid = self.crs_uuid)



"""High level local depth adjustment function."""

import logging

log = logging.getLogger(__name__)

import os
import math as maths
import numpy as np

import resqpy.model as rq
import resqpy.olio.xml_et as rqet

from resqpy.derived_model._common import _displacement_properties, _prepare_simple_inheritance, _write_grid, _establish_model_and_source_grid
from resqpy.derived_model._copy_grid import copy_grid


def local_depth_adjustment(epc_file,
                           source_grid,
                           centre_x,
                           centre_y,
                           radius,
                           centre_shift,
                           use_local_coords,
                           decay_shape = 'quadratic',
                           ref_k0 = 0,
                           store_displacement = False,
                           inherit_properties = False,
                           inherit_realization = None,
                           inherit_all_realizations = False,
                           new_grid_title = None,
                           new_epc_file = None):
    """Applies a local depth adjustment to the grid, adding as a new grid part in the model.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): a multi-layer RESQML grid object; if None, the epc_file is loaded
          and it should contain one ijk grid object (or one 'ROOT' grid) which is used as the source grid
       centre_x, centre_y (floats): the centre of the depth adjustment, corresponding to the location of maximum change
          in depth; crs is implicitly that of the grid but see also use_local_coords argument
       radius (float): the radius of adjustment of depths; units are implicitly xy (projected) units of grid crs
       centre_shift (float): the maximum vertical depth adjustment; units are implicily z (vertical) units of grid crs;
          use positive value to increase depth, negative to make shallower
       use_local_coords (boolean): if True, centre_x & centre_y are taken to be in the local coordinates of the grid's
          crs; otherwise the global coordinates
       decay_shape (string): 'linear' yields a cone shaped change in depth values; 'quadratic' (the default) yields a
          bell shaped change
       ref_k0 (integer, default 0): the layer in the grid to use as reference for determining the distance of a pillar
          from the centre of the depth adjustment; the corners of the top face of the reference layer are used
       store_displacement (boolean, default False): if True, 3 grid property parts are created, one each for x, y, & z
          displacement of cells' centres brought about by the local depth shift
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the adjusted grid (& crs)

    returns:
       new grid object which is a copy of the source grid with the local depth adjustment applied
    """

    log.info('adjusting depth')
    log.debug('centre x: {0:3.1f}; y: {1:3.1f}'.format(centre_x, centre_y))
    if use_local_coords:
        log.debug('centre x & y interpreted in local crs')
    log.debug('radius of influence: {0:3.1f}'.format(radius))
    log.debug('depth shift at centre: {0:5.3f}'.format(centre_shift))
    log.debug('decay shape: ' + decay_shape)
    log.debug('reference layer (k0 protocol): ' + str(ref_k0))

    assert epc_file or new_epc_file, 'epc file name not specified'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    model, source_grid = _establish_model_and_source_grid(epc_file, source_grid)
    assert source_grid.grid_representation == 'IjkGrid'
    assert model is not None

    # take a copy of the grid
    grid = copy_grid(source_grid, model)

    # if not use_local_coords, convert centre_x & y into local_coords
    crs_root = grid.extract_crs_root()
    assert (crs_root is not None)
    if not use_local_coords:
        rotation = abs(float(rqet.node_text(rqet.find_tag(crs_root, 'ArealRotation'))))
        if rotation > 0.001:
            log.error('unable to account for rotation in crs: use local coordinates')
            return
        centre_x -= float(rqet.node_text(rqet.find_tag(crs_root, 'XOffset')))
        centre_y -= float(rqet.node_text(rqet.find_tag(crs_root, 'YOffset')))
    z_inc_down = rqet.bool_from_text(rqet.node_text(rqet.find_tag(crs_root, 'ZIncreasingDownward')))

    if not z_inc_down:
        centre_shift = -centre_shift

    # cache geometry in memory; needed prior to writing new coherent set of arrays to hdf5
    grid.cache_all_geometry_arrays()
    if grid.has_split_coordinate_lines:
        reshaped_points = grid.points_cached.copy()
    else:
        nkp1, njp1, nip1, xyz = grid.points_cached.shape
        reshaped_points = grid.points_cached.copy().reshape((nkp1, njp1 * nip1, xyz))
    assert reshaped_points.ndim == 3 and reshaped_points.shape[2] == 3
    assert ref_k0 >= 0 and ref_k0 < reshaped_points.shape[0]

    log.debug('reshaped_points.shape: ' + str(reshaped_points.shape))

    log.debug('min z before depth adjustment: ' + str(np.nanmin(reshaped_points[:, :, 2])))

    # for each pillar, find x, y for k = reference_layer_k0
    pillars_adjusted = 0

    # todo: replace with numpy array operations
    radius_sqr = radius * radius
    for pillar in range(reshaped_points.shape[1]):
        x, y, z = tuple(reshaped_points[ref_k0, pillar, :])
        # find distance of this pillar from the centre
        dx = centre_x - x
        dy = centre_y - y
        distance_sqr = (dx * dx) + (dy * dy)
        # if this pillar is beyond radius of influence, no action needed
        if distance_sqr > radius_sqr:
            continue
        distance = maths.sqrt(distance_sqr)
        # compute decayed shift as function of distance
        shift = _decayed_shift(centre_shift, distance, radius, decay_shape)
        # adjust depth values for pillar in cached array
        log.debug('adjusting pillar number {0} at x: {1:3.1f}, y: {2:3.1f}, distance: {3:3.1f} by {4:5.3f}'.format(
            pillar, x, y, distance, shift))
        reshaped_points[:, pillar, 2] += shift
        pillars_adjusted += 1

    # if no pillars adjusted: warn and return
    if pillars_adjusted == 0:
        log.warning('no pillars adjusted')
        return

    log.debug('min z after depth adjustment: ' + str(np.nanmin(reshaped_points[:, :, 2])))
    if grid.has_split_coordinate_lines:
        grid.points_cached[:] = reshaped_points
    else:
        grid.points_cached[:] = reshaped_points.reshape((nkp1, njp1, nip1, xyz))


#   model.copy_part(old_uuid, grid.uuid, change_hdf5_refs = True)   # copies the xml, substituting the new uuid in the root node (and in hdf5 refs)
    log.info(str(pillars_adjusted) + ' pillars adjusted')

    # build cell displacement property array(s)
    if store_displacement:
        log.debug('generating cell displacement property arrays')
        displacement_collection = _displacement_properties(grid, source_grid)
    else:
        displacement_collection = None

    collection = _prepare_simple_inheritance(grid, source_grid, inherit_properties, inherit_realization,
                                             inherit_all_realizations)
    if collection is None:
        collection = displacement_collection
    elif displacement_collection is not None:
        collection.inherit_imported_list_from_other_collection(displacement_collection, copy_cached_arrays = False)

    if new_grid_title is None or len(new_grid_title) == 0:
        new_grid_title = 'grid derived from {0} with local depth shift of {1:3.1f} applied'.format(
            str(rqet.citation_title_for_node(source_grid.root)), centre_shift)

    # write model
    model.h5_release()
    if new_epc_file:
        _write_grid(new_epc_file, grid, property_collection = collection, grid_title = new_grid_title, mode = 'w')
    else:
        ext_uuid, _ = model.h5_uuid_and_path_for_node(rqet.find_nested_tags(source_grid.root, ['Geometry', 'Points']),
                                                      'Coordinates')
        _write_grid(epc_file,
                    grid,
                    ext_uuid = ext_uuid,
                    property_collection = collection,
                    grid_title = new_grid_title,
                    mode = 'a')

    return grid


def _decayed_shift(centre_shift, distance, radius, decay_shape):
    norm_dist = min(distance / radius, 1.0)  # 0..1
    if decay_shape == 'linear':
        return (1.0 - norm_dist) * centre_shift
    elif decay_shape == 'quadratic':
        if norm_dist >= 0.5:
            x = (1.0 - norm_dist)
            return 2.0 * x * x * centre_shift
        else:
            return centre_shift * (1.0 - 2.0 * norm_dist * norm_dist)
    else:
        raise ValueError('unrecognized decay shape: ' + decay_shape)



""" Small utility functions related to use of pure binary files """

version = '4th September 2020'

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import numpy as np

ab_dtype_dict = {'.db': np.float64, '.fb': np.float32, '.lb': np.int64, '.ib': np.int32, '.bb': np.int8}


def load_array_from_ab_file(file_name, shape, return_64_bit = False):
    """Loads a pure binary file into a numpy array, optionally converting to 64 bit."""

    count = 1
    for axis in range(len(shape)):
        count *= shape[axis]
    dtype = ab_dtype_dict[file_name[-3:]]
    with open(file_name, 'rb') as fp:
        a = np.fromfile(fp, dtype = dtype, count = count).reshape(tuple(shape))
    try:  # expected to return null
        c = fp.read(1)
        if len(c):
            log.warning('binary file contains more data than expected: ' + file_name)
    except Exception:
        pass
    return a


def cp_binary_filename(file_name, nexus_ordering = True):
    """Returns a version of the file name with extension adjusted to indicate reseq order and pure binary."""

    if file_name[-9:] == '.reseq.db':
        root_name = file_name[:-9]
    elif file_name[-3:] == '.db':
        root_name = file_name[:-3]
    else:
        root_name = file_name
    if nexus_ordering:
        return root_name + '.db'
    else:
        return root_name + '.reseq.db'


def binary_file_extension_and_np_type_for_data_type(data_type):
    """Returns a file extension suitable for a pure binary array (ab) file of given data type."""

    if data_type in ['real', 'float']:
        ab_type = np.dtype('f8')  # 8 byte floating point (-double in ab_* suite)
        extension = '.db'
    elif data_type in ['int', 'integer']:
        ab_type = np.dtype('i8')  # NB: 8 byte integers not currently supported by ab_* suite
        extension = '.lb'
    elif data_type in ['bool', 'boolean']:
        ab_type = np.dtype('?')  # single byte boolean (-byte in ab_* suite)
        extension = '.bb'
    else:
        log.error('Unknown data_type [%s] passed to load_array_from_file()', data_type)
        assert (False)
    return (extension, ab_type)



# resqpy test module including well blocking of trajectories against challenging grid geometries

version = '29th April 2021'

import logging

log = logging.getLogger(__name__)

import math as maths
import os

import numpy as np
import pandas as pd

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.model as rq
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.olio.xml_et as rqet
import resqpy.rq_import as rqi
import resqpy.well as rqw


def test_s_bend_fn(tmp_path, epc = None):

    if epc is None:
        # use pytest temporary directory fixture
        # https://docs.pytest.org/en/stable/tmpdir.html
        epc = str(os.path.join(tmp_path, f"{bu.new_uuid()}.epc"))

    # create s-bend grid

    nk = 5
    nj = 12
    ni_tail = 5
    ni_bend = 18
    ni_half_mid = 2
    ni = 2 * (ni_tail + ni_bend + ni_half_mid)

    total_thickness = 12.0
    layer_thickness = total_thickness / float(nk)
    flat_dx_di = 10.0
    horst_dx_dk = 0.25 * flat_dx_di / float(nk)
    horst_dz = 1.73 * layer_thickness
    horst_half_dx = horst_dx_dk * horst_dz / layer_thickness
    dy_dj = 8.0
    top_depth = 100.0

    assert ni_bend % 2 == 0, 'ni_bend must be even for horizontal faulting'
    assert ni_tail >= 5, 'ni_tail must be at least 5 for horst blocks'

    bend_theta_di = maths.pi / float(ni_bend)
    outer_radius = 2.0 * total_thickness

    bend_a_centre_xz = (flat_dx_di * float(ni_tail), top_depth + outer_radius)
    bend_b_centre_xz = (flat_dx_di * float(ni_tail - 2.0 * ni_half_mid),
                        top_depth + 3.0 * outer_radius - total_thickness)

    points = np.empty((nk + 1, nj + 1, ni + 1, 3))

    for k in range(nk + 1):
        if k == nk // 2 + 1:
            points[k] = points[k - 1]  # pinched out layer
        else:
            for i in range(ni + 1):
                if i < ni_tail + 1:
                    x = flat_dx_di * float(i)
                    z = top_depth + float(k) * layer_thickness  # will introduce a thick layer after pinchout
                elif i < ni_tail + ni_bend:
                    theta = (i - ni_tail) * bend_theta_di
                    radius = outer_radius - float(k) * layer_thickness
                    x = bend_a_centre_xz[0] + radius * maths.sin(theta)
                    z = bend_a_centre_xz[1] - radius * maths.cos(theta)
                elif i < ni_tail + ni_bend + 2 * ni_half_mid + 1:
                    x = flat_dx_di * float(ni_tail - (i - (ni_tail + ni_bend)))
                    z = top_depth + 2.0 * outer_radius - float(k) * layer_thickness
                elif i < ni_tail + 2 * ni_bend + 2 * ni_half_mid:
                    theta = (i - (ni_tail + ni_bend + 2 * ni_half_mid)) * bend_theta_di
                    radius = outer_radius - float(nk - k) * layer_thickness
                    x = bend_b_centre_xz[0] - radius * maths.sin(theta)
                    z = bend_b_centre_xz[1] - radius * maths.cos(theta)
                else:
                    x = flat_dx_di * float((i - (ni - ni_tail)) + ni_tail - 2 * ni_half_mid)
                    if i == ni - 1 or i == ni - 4:
                        x += horst_dx_dk * float(k)
                    elif i == ni - 2 or i == ni - 3:
                        x -= horst_dx_dk * float(k)
                    z = top_depth + 4.0 * outer_radius + float(k) * layer_thickness - 2.0 * total_thickness
                points[k, :, i] = (x, 0.0, z)

    for j in range(nj + 1):
        points[:, j, :, 1] = dy_dj * float(j)

    model = rq.Model(epc_file = epc, new_epc = True, create_basics = True, create_hdf5_ext = True)

    grid = grr.Grid(model)

    crs = rqc.Crs(model)
    crs_node = crs.create_xml()
    if model.crs_root is None:
        model.crs_root = crs_node

    grid.grid_representation = 'IjkGrid'
    grid.extent_kji = np.array((nk, nj, ni), dtype = 'int')
    grid.nk, grid.nj, grid.ni = nk, nj, ni
    grid.k_direction_is_down = True  # dominant layer direction, or paleo-direction
    grid.pillar_shape = 'straight'
    grid.has_split_coordinate_lines = False
    grid.k_gaps = None
    grid.crs_uuid = crs.uuid
    grid.crs_root = crs_node

    grid.points_cached = points

    grid.geometry_defined_for_all_pillars_cached = True
    grid.geometry_defined_for_all_cells_cached = True
    grid.grid_is_right_handed = crs.is_right_handed_xyz()

    grid.write_hdf5_from_caches()
    grid.create_xml()

    # create a well trajectory and md datum

    def df_trajectory(x, y, z):
        N = len(x)
        assert len(y) == N and len(z) == N
        df = pd.DataFrame(columns = ['MD', 'X', 'Y', 'Z'])
        md = np.zeros(N)
        for n in range(N - 1):
            md[n + 1] = md[n] + vec.naive_length((x[n + 1] - x[n], y[n + 1] - y[n], z[n + 1] - z[n]))
        df.MD = md
        df.X = x
        df.Y = y
        df.Z = z
        return df

    x = np.array(
        [0.0, flat_dx_di * float(ni_tail) + outer_radius, flat_dx_di * (float(ni_tail) - 0.5), 0.0, -outer_radius])
    y = np.array([0.0, dy_dj * 0.5, dy_dj * float(nj) / 2.0, dy_dj * (float(nj) - 0.5), dy_dj * float(nj)])
    z = np.array([
        0.0, top_depth - total_thickness, top_depth + 2.0 * outer_radius - total_thickness / 2.0,
        top_depth + 3.0 * outer_radius - total_thickness, top_depth + 4.0 * outer_radius
    ])

    df = df_trajectory(x, y, z)

    datum = rqw.MdDatum(model, crs_uuid = crs.uuid, location = (x[0], y[0], z[0]))
    datum.create_xml()

    trajectory = rqw.Trajectory(model, md_datum = datum, data_frame = df, length_uom = 'm', well_name = 'ANGLED_WELL')

    rqc.Crs(model, uuid = rqet.uuid_for_part_root(trajectory.crs_root)).uuid == crs.uuid

    trajectory.write_hdf5()
    trajectory.create_xml()

    # add more wells

    x = np.array([
        0.0, flat_dx_di * float(ni_tail), flat_dx_di * 2.0 * float(ni_tail - ni_half_mid) + outer_radius, -outer_radius
    ])
    y = np.array([0.0, dy_dj * float(nj) * 0.59, dy_dj * 0.67, dy_dj * 0.5])
    z = np.array([
        0.0, top_depth - total_thickness, top_depth + 4.0 * outer_radius - 1.7 * total_thickness,
        top_depth + 4.0 * outer_radius - 1.7 * total_thickness
    ])

    df = df_trajectory(x, y, z)

    traj_2 = rqw.Trajectory(model, md_datum = datum, data_frame = df, length_uom = 'm', well_name = 'HORST_WELL')
    traj_2.write_hdf5()
    traj_2.create_xml()
    traj_2.control_points

    x = np.array([0.0, 0.0, 0.0])
    y = np.array([0.0, dy_dj * float(nj) * 0.53, dy_dj * float(nj) * 0.53])
    z = np.array([0.0, top_depth - total_thickness, top_depth + 4.0 * outer_radius])

    df = df_trajectory(x, y, z)

    traj_3 = rqw.Trajectory(model, md_datum = datum, data_frame = df, length_uom = 'm', well_name = 'VERTICAL_WELL')
    traj_3.write_hdf5()
    traj_3.create_xml()
    traj_3.control_points

    n_x = flat_dx_di * float(ni_tail) * 0.48
    n_y = dy_dj * float(nj) / 9.1
    o_y = -dy_dj * 0.45
    nd_x = n_y / 3.0
    x = np.array([
        0.0, n_x, n_x, n_x + nd_x, n_x + 2.0 * nd_x, n_x + 3.0 * nd_x, n_x + 4.0 * nd_x, n_x + 5.0 * nd_x,
        n_x + 6.0 * nd_x, n_x + 7.0 * nd_x, n_x + 8.0 * nd_x, n_x + 8.0 * nd_x
    ])
    y = np.array([
        0.0, o_y, o_y + n_y, o_y + 2.0 * n_y, o_y + 3.0 * n_y, o_y + 4.0 * n_y, o_y + 5.0 * n_y, o_y + 6.0 * n_y,
        o_y + 7.0 * n_y, o_y + 8.0 * n_y, o_y + 9.0 * n_y, o_y + 10.0 * n_y
    ])
    n_z1 = top_depth + total_thickness * 0.82
    n_z2 = top_depth - total_thickness * 0.17
    z = np.array([0.0, n_z1, n_z1, n_z2, n_z2, n_z1, n_z1, n_z2, n_z2, n_z1, n_z1, n_z2])

    df = df_trajectory(x, y, z)

    traj_4 = rqw.Trajectory(model, md_datum = datum, data_frame = df, length_uom = 'm', well_name = 'NESSIE_WELL')
    traj_4.write_hdf5()
    traj_4.create_xml()
    traj_4.control_points

    # block wells against grid geometry

    log.info('unfaulted grid blocking of well ' + str(rqw.well_name(trajectory)))
    bw = rqw.BlockedWell(model, grid = grid, trajectory = trajectory)
    bw.write_hdf5()
    bw.create_xml()
    assert bw.cell_count == 19

    log.info('unfaulted grid blocking of well ' + str(rqw.well_name(traj_2)))
    bw_2 = rqw.BlockedWell(model, grid = grid, trajectory = traj_2)
    bw_2.write_hdf5()
    bw_2.create_xml()
    assert bw_2.cell_count == 33

    log.info('unfaulted grid blocking of well ' + str(rqw.well_name(traj_3)))
    bw_3 = rqw.BlockedWell(model, grid = grid, trajectory = traj_3)
    bw_3.write_hdf5()
    bw_3.create_xml()
    assert bw_3.cell_count == 18

    log.info('unfaulted grid blocking of well ' + str(rqw.well_name(traj_4)))
    bw_4 = rqw.BlockedWell(model, grid = grid, trajectory = traj_4)
    bw_4.write_hdf5()
    bw_4.create_xml()
    assert bw_4.cell_count == 26

    # derive a faulted version of the grid

    cp = grid.corner_points(cache_cp_array = True).copy()

    # IK plane faults
    cp[:, 3:, :, :, :, :, :] += (flat_dx_di * 0.7, 0.0, layer_thickness * 1.3)
    cp[:, 5:, :, :, :, :, :] += (flat_dx_di * 0.4, 0.0, layer_thickness * 0.9)
    cp[:, 8:, :, :, :, :, :] += (flat_dx_di * 0.3, 0.0, layer_thickness * 0.6)

    # JK plane faults
    cp[:, :, ni_tail + ni_bend // 2:, :, :, :, 0] -= flat_dx_di * 0.57  # horizontal break mid top bend
    cp[:, :, ni_tail + ni_bend + ni_half_mid:, :, :, :, 2] += layer_thickness * 1.27  # vertical break in mid section

    # zig-zag fault
    j_step = nj // (ni_tail - 2)
    for i in range(ni_tail - 1):
        j_start = i * j_step
        if j_start >= nj:
            break
        cp[:, j_start:, i, :, :, :, 2] += 1.1 * total_thickness

    # JK horst blocks
    cp[:, :, ni - 4, :, :, :, :] -= (horst_half_dx, 0.0, horst_dz)
    cp[:, :, ni - 3:, :, :, :, 0] -= 2.0 * horst_half_dx
    cp[:, :, ni - 2, :, :, :, :] += (-horst_half_dx, 0.0, horst_dz)
    cp[:, :, ni - 1:, :, :, :, 0] -= 2.0 * horst_half_dx

    # JK horst block mid lower bend
    bend_horst_dz = horst_dz * maths.tan(bend_theta_di)
    cp[:, :, ni - (ni_tail + ni_bend // 2 + 1):ni - (ni_tail + ni_bend // 2 - 1), :, :, :, :] -= (horst_dz, 0.0,
                                                                                                  bend_horst_dz)
    cp[:, :, ni - (ni_tail + ni_bend // 2 - 1):, :, :, :, 2] -= 2.0 * bend_horst_dz

    faulted_grid = rqi.grid_from_cp(model,
                                    cp,
                                    crs.uuid,
                                    max_z_void = 0.01,
                                    split_pillars = True,
                                    split_tolerance = 0.01,
                                    ijk_handedness = 'right' if grid.grid_is_right_handed else 'left',
                                    known_to_be_straight = True)

    faulted_grid.write_hdf5_from_caches()
    faulted_grid.create_xml()

    # block wells against faulted grid

    log.info('faulted grid blocking of well ' + str(rqw.well_name(trajectory)))
    fbw = rqw.BlockedWell(model, grid = faulted_grid, trajectory = trajectory)
    fbw.write_hdf5()
    fbw.create_xml()
    assert fbw.cell_count == 32

    log.info('faulted grid blocking of well ' + str(rqw.well_name(traj_2)))
    fbw_2 = rqw.BlockedWell(model, grid = faulted_grid, trajectory = traj_2)
    fbw_2.write_hdf5()
    fbw_2.create_xml()
    assert fbw_2.cell_count == 26

    log.info('faulted grid blocking of well ' + str(rqw.well_name(traj_3)))
    fbw_3 = rqw.BlockedWell(model, grid = faulted_grid, trajectory = traj_3)
    fbw_3.write_hdf5()
    fbw_3.create_xml()
    assert fbw_3.cell_count == 14

    log.info('faulted grid blocking of well ' + str(rqw.well_name(traj_4)))
    fbw_4 = rqw.BlockedWell(model, grid = faulted_grid, trajectory = traj_4)
    fbw_4.write_hdf5()
    fbw_4.create_xml()
    assert fbw_4.cell_count == 16

    # create a version of the faulted grid with a k gap

    k_gap_grid = rqi.grid_from_cp(model,
                                  cp,
                                  crs.uuid,
                                  max_z_void = 0.01,
                                  split_pillars = True,
                                  split_tolerance = 0.01,
                                  ijk_handedness = 'right' if grid.grid_is_right_handed else 'left',
                                  known_to_be_straight = True)

    # convert second layer to a K gap
    k_gap_grid.nk -= 1
    k_gap_grid.extent_kji[0] = k_gap_grid.nk
    k_gap_grid.k_gaps = 1
    k_gap_grid.k_gap_after_array = np.zeros(k_gap_grid.nk - 1, dtype = bool)
    k_gap_grid.k_gap_after_array[0] = True
    k_gap_grid.k_raw_index_array = np.zeros(k_gap_grid.nk, dtype = int)
    for k in range(1, k_gap_grid.nk):
        k_gap_grid.k_raw_index_array[k] = k + 1

    # clear some attributes which may no longer be valid
    k_gap_grid.pinchout = None
    k_gap_grid.inactive = None
    k_gap_grid.grid_skin = None
    if hasattr(k_gap_grid, 'array_thickness'):
        delattr(k_gap_grid, 'array_thickness')

    k_gap_grid.write_hdf5_from_caches()
    k_gap_grid.create_xml()

    k_gap_grid_uuid = k_gap_grid.uuid

    # reload k gap grid object to ensure it is properly initialised

    k_gap_grid = None
    k_gap_grid = grr.Grid(model, uuid = k_gap_grid_uuid)

    # block wells against faulted grid with k gap

    log.info('k gap grid blocking of well ' + str(rqw.well_name(trajectory)))
    try:
        gbw = rqw.BlockedWell(model, grid = k_gap_grid, trajectory = trajectory)
        gbw.write_hdf5()
        gbw.create_xml()
        assert gbw.cell_count == 24
    except Exception:
        log.exception('failed to block well against k gap grid')

    log.info('k gap grid blocking of well ' + str(rqw.well_name(traj_2)))
    try:
        gbw_2 = rqw.BlockedWell(model, grid = k_gap_grid, trajectory = traj_2)
        gbw_2.write_hdf5()
        gbw_2.create_xml()
        assert gbw_2.cell_count == 20
    except Exception:
        log.exception('failed to block well against k gap grid')

    log.info('k gap grid blocking of well ' + str(rqw.well_name(traj_3)))
    try:
        gbw_3 = rqw.BlockedWell(model, grid = k_gap_grid, trajectory = traj_3)
        gbw_3.write_hdf5()
        gbw_3.create_xml()
        assert gbw_3.cell_count == 10
    except Exception:
        log.exception('failed to block well against k gap grid')

    log.info('k gap grid blocking of well ' + str(rqw.well_name(traj_4)))
    try:
        gbw_4 = rqw.BlockedWell(model, grid = k_gap_grid, trajectory = traj_4)
        gbw_4.write_hdf5()
        gbw_4.create_xml()
        assert gbw_4.cell_count == 10
    except Exception:
        log.exception('failed to block well against k gap grid')

    # store model

    model.store_epc()

    assert k_gap_grid.k_gaps
    assert len(k_gap_grid.k_raw_index_array) == k_gap_grid.nk
    assert len(k_gap_grid.k_gap_after_array) == k_gap_grid.nk - 1
    assert k_gap_grid.pinched_out((1, 0, 2))

    # clean up
    model.h5_release()
    os.remove(model.h5_file_name())



"""combined_surface.py: surface class based on resqml standard."""

version = '4th November 2021'

# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company
# GOCAD is also a trademark of Emerson

import logging

log = logging.getLogger(__name__)

import numpy as np


class CombinedSurface:
    """Class allowing a collection of Surface objects to be treated as a single surface.
    
    Not a RESQML class in its own right.
    """

    def __init__(self, surface_list, crs_uuid = None):
        """Initialise a CombinedSurface object from a list of Surface (and/or CombinedSurface) objects.

        arguments:
           surface_list (list of Surface and/or CombinedSurface objects): the new object is the combination of these surfaces
           crs_uuid (uuid.UUID, optional): if present, all contributing surfaces must refer to this crs

        note:
           all contributing surfaces should be established before initialising this object;
           all contributing surfaces must refer to the same crs; this class of object is not part of the RESQML
           standard and cannot be saved in a RESQML dataset - it is a high level derived object class
        """

        assert len(surface_list) > 0
        self.surface_list = surface_list
        self.crs_uuid = crs_uuid
        if self.crs_uuid is None:
            self.crs_uuid = surface_list[0].crs_uuid
        self.patch_count_list = []
        self.triangle_count_list = []
        self.points_count_list = []
        self.is_combined_list = []
        self.triangles = None
        self.points = None
        for surface in surface_list:
            is_combined = isinstance(surface, CombinedSurface)
            self.is_combined_list.append(is_combined)
            if is_combined:
                self.patch_count_list.append(sum(surface.patch_count_list))
            else:
                self.patch_count_list.append(len(surface.patch_list))
            t, p = surface.triangles_and_points()
            self.triangle_count_list.append(len(t))
            self.points_count_list.append(len(p))

    def surface_index_for_triangle_index(self, tri_index):
        """Return the index of the surface containing the triangle and local triangle index.
        
        Arguments:
            tri_index: triangle index in the combined surface
        """
        for s_i in range(len(self.surface_list)):
            if tri_index < self.triangle_count_list[s_i]:
                return s_i, tri_index
            tri_index -= self.triangle_count_list[s_i]
        return None

    def triangles_and_points(self):
        """Returns the composite triangles and points for the combined surface."""

        if self.triangles is None:
            points_offset = 0
            for surface in self.surface_list:
                (t, p) = surface.triangles_and_points()
                if points_offset == 0:
                    self.triangles = t.copy()
                    self.points = p.copy()
                else:
                    self.triangles = np.concatenate((self.triangles, t.copy() + points_offset))
                    self.points = np.concatenate((self.points, p.copy()))
                points_offset += p.shape[0]

        return self.triangles, self.points



from resqpy.time_series._any_time_series import AnyTimeSeries
import uuid
import pytest


def test_is_equivalent_other_ts_is_none(example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    # act
    result = any_time_series.is_equivalent(None)
    # assert
    assert result is False


def test_is_equivalent_self_is_other_ts(example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    # act
    result = any_time_series.is_equivalent(any_time_series)
    # assert
    assert result is True


def test_is_equivalent_matching_uuids(example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    test_time_series = AnyTimeSeries(model)
    test_uuid = uuid.uuid4()
    any_time_series.uuid = test_uuid
    test_time_series.uuid = test_uuid
    # act
    result = any_time_series.is_equivalent(test_time_series)
    # assert
    assert result is True


def test_is_equivalent_none_matching_uuids_none_matching_timestamps(example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    test_time_series = AnyTimeSeries(model)
    test_uuid = uuid.uuid4()
    any_time_series.uuid = test_uuid
    test_time_series.uuid = uuid.uuid4()
    any_time_series.timestamps = [1, 2]
    test_time_series.timestamps = [1, 2, 3]
    # act
    result = any_time_series.is_equivalent(test_time_series)
    # assert
    assert result is False


def test_is_equivalent_matching_uuids_matching_timestamps_none_matching_timeframe(example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    test_time_series = AnyTimeSeries(model)
    test_uuid = uuid.uuid4()
    any_time_series.uuid = test_uuid
    test_time_series.uuid = uuid.uuid4()
    any_time_series.timestamps = [1, 2]
    test_time_series.timestamps = [1, 2]
    timeframe = 'test'
    test_timeframe = 'test_1'
    any_time_series.timeframe = timeframe
    test_time_series.timeframe = test_timeframe
    # act
    result = any_time_series.is_equivalent(test_time_series)
    # assert
    assert result is False


def test_is_equivalent_matching_uuids_matching_timestamps_matching_timeframe(example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    test_time_series = AnyTimeSeries(model)
    test_uuid = uuid.uuid4()
    any_time_series.uuid = test_uuid
    test_time_series.uuid = uuid.uuid4()
    any_time_series.timestamps = [1, 2]
    test_time_series.timestamps = [1, 2]
    timeframe = 'test'
    test_timeframe = timeframe
    any_time_series.timeframe = timeframe
    test_time_series.timeframe = test_timeframe
    # act
    result = any_time_series.is_equivalent(test_time_series)
    # assert
    assert result is None


@pytest.mark.parametrize('timestamp', [['2014-01-01T00:00:00.588Z']])
def test_timestamp(timestamp, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    any_time_series.timestamps = timestamp
    # act
    result = any_time_series.timestamp(1)
    # assert
    assert result is None


@pytest.mark.parametrize('timestamp', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_valid_timestamp(timestamp, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    any_time_series.timestamps = timestamp
    # act
    result = any_time_series.timestamp(0)
    # assert
    assert result == '2014-01-01T00:00:00.588Z'


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_itr_timestamp(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.iter_timestamps()
    # assert
    assert list(result) == ['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_timestamp(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.index_for_timestamp('2014-01-01T00:00:00.588Z')
    # assert
    assert result == 0


@pytest.mark.parametrize('timestamps', [['2014-01-01T00:00:00.588Z', '2015-02-02T00:00:00.588Z']])
def test_index_timestamp_not_present(timestamps, example_model_and_crs):
    # arrange
    model, crs = example_model_and_crs
    any_time_series = AnyTimeSeries(model)
    any_time_series.timestamps = timestamps
    # act
    result = any_time_series.index_for_timestamp('2016-01-01T00:00:00.588Z')
    # assert
    assert result is None



import os

import math as maths
import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_almost_equal

import resqpy.crs as rqc
import resqpy.derived_model as rqdm
import resqpy.fault as rqf
import resqpy.grid as grr
import resqpy.lines as rql
import resqpy.model as rq
import resqpy.olio.box_utilities as bx
import resqpy.olio.fine_coarse as rqfc
import resqpy.olio.uuid as bu
import resqpy.property as rqp
import resqpy.surface as rqs

import resqpy.well as rqw

from resqpy.olio.random_seed import seed

seed(2349857)


def test_add_single_cell_grid(tmp_path):

    epc = os.path.join(tmp_path, 'amoeba.epc')

    points = np.array([(100.0, 250.0, -3500.0), (140.0, 200.0, -3700.0), (300.0, 400.0, -3600.0),
                       (180.0, 300.0, -3800.0), (220.0, 350.0, -3750.0)])
    expected_xyz_box = np.array([(100.0, 200.0, -3800.0), (300.0, 400.0, -3500.0)])

    # create a single cell grid containing points
    rqdm.add_single_cell_grid(points, new_grid_title = 'Amoeba', new_epc_file = epc)

    # re-open model and have a quick look at the grid
    model = rq.Model(epc)
    assert model is not None
    grid = grr.Grid(model, uuid = model.uuid(title = 'Amoeba'))
    assert grid is not None
    assert tuple(grid.extent_kji) == (1, 1, 1)
    assert_array_almost_equal(grid.xyz_box(lazy = False), expected_xyz_box)


def test_add_zone_by_layer_property(tmp_path):

    def check_zone_prop(z_prop):
        assert z_prop is not None
        assert not z_prop.is_continuous()
        assert not z_prop.is_points()
        assert z_prop.indexable_element() == 'layers'
        lpk_uuid = z_prop.local_property_kind_uuid()
        assert lpk_uuid is not None
        lpk = rqp.PropertyKind(z_prop.model, uuid = lpk_uuid)
        assert lpk.title == 'zone'

    epc = os.path.join(tmp_path, 'in the zone.epc')

    model = rq.new_model(epc)

    # create a basic block grid
    grid = grr.RegularGrid(model, extent_kji = (4, 3, 2), title = 'In The Zone', set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    grid_uuid = grid.uuid

    model.store_epc()

    # add zone property based on an explicit vector (one value per layer)
    zone_vector = (2, 7, 5, 7)
    v, z_uuid = rqdm.add_zone_by_layer_property(epc_file = epc,
                                                zone_by_layer_vector = (2, 7, 5, 7),
                                                title = 'from vector')
    assert tuple(v) == zone_vector

    # check that zone property looks okay
    model = rq.Model(epc)
    z_prop = rqp.Property(model, uuid = z_uuid)
    check_zone_prop(z_prop)

    # add a neatly set up grid cells property
    za = np.array((1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 5, 5, 5, 5, 5, 5),
                  dtype = int).reshape(grid.extent_kji)
    za_uuid = rqdm.add_one_grid_property_array(epc,
                                               za,
                                               discrete = True,
                                               property_kind = 'code',
                                               title = 'clean zone',
                                               grid_uuid = grid_uuid,
                                               null_value = -1)
    assert za_uuid is not None

    # add a zone by layer property based on the neat cells property
    v, z_uuid = rqdm.add_zone_by_layer_property(epc_file = epc,
                                                zone_by_cell_property_uuid = za_uuid,
                                                title = 'from cells array')
    assert tuple(v) == (1, 2, 3, 5)

    # check that zone property looks okay
    model = rq.Model(epc)
    z_prop = rqp.Property(model, uuid = z_uuid)
    check_zone_prop(z_prop)

    # make the cells array less tidy and add another copy
    za[1, 2, :] = 3
    za_uuid = rqdm.add_one_grid_property_array(epc,
                                               za,
                                               discrete = True,
                                               property_kind = 'code',
                                               title = 'messy zone',
                                               grid_uuid = grid_uuid,
                                               null_value = -1)
    assert za_uuid is not None

    # fail to add a zone by layer property based on the messy cells property
    with pytest.raises(Exception):
        v, z2_uuid = rqdm.add_zone_by_layer_property(epc_file = epc,
                                                     zone_by_cell_property_uuid = za_uuid,
                                                     use_dominant_zone = False,
                                                     title = 'should fail')

    # add a zone by layer property based on the neat cells property
    v, z3_uuid = rqdm.add_zone_by_layer_property(epc_file = epc,
                                                 zone_by_cell_property_uuid = za_uuid,
                                                 use_dominant_zone = True,
                                                 title = 'from messy cells array')
    assert tuple(v) == (1, 2, 3, 5)

    # check that zone property looks okay
    model = rq.Model(epc)
    grid = model.grid()
    z_prop = rqp.Property(model, uuid = z3_uuid)
    check_zone_prop(z_prop)

    # create a zonal grid based on a (neat) zone property array
    z_grid = rqdm.zonal_grid(epc,
                             source_grid = grid,
                             zone_uuid = z_uuid,
                             use_dominant_zone = False,
                             inactive_laissez_faire = True,
                             new_grid_title = 'zonal grid')
    assert z_grid is not None
    assert z_grid.nk == 4

    # and another zonal grid based on the dominant zone
    z3_grid = rqdm.zonal_grid(epc,
                              source_grid = grid,
                              zone_uuid = za_uuid,
                              use_dominant_zone = True,
                              new_grid_title = 'dominant zone grid')
    assert z3_grid is not None
    assert z3_grid.nk == 4


def test_single_layer_grid(tmp_path):

    epc = os.path.join(tmp_path, 'squash.epc')

    model = rq.new_model(epc)

    # create a basic block grid with geometry
    grid = grr.RegularGrid(model,
                           extent_kji = (4, 3, 2),
                           origin = (1000.0, 2000.0, 3000.0),
                           dxyz = (100.0, 130.0, 25.0),
                           title = 'to be squashed',
                           set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    grid_uuid = grid.uuid
    model.store_epc()

    # create a single layer version of the grid
    simplified = rqdm.single_layer_grid(epc, source_grid = grid, new_grid_title = 'squashed')
    assert simplified is not None
    simplified_uuid = simplified.uuid

    # re-open the model and load the new grid
    model = rq.Model(epc)
    s_uuid = model.uuid(obj_type = 'IjkGridRepresentation', title = 'squashed')
    assert bu.matching_uuids(s_uuid, simplified_uuid)
    simplified = grr.any_grid(model, uuid = s_uuid)
    assert simplified.nk == 1
    simplified.cache_all_geometry_arrays()
    assert not simplified.has_split_coordinate_lines
    assert simplified.points_cached.shape == (2, 4, 3, 3)
    assert_array_almost_equal(simplified.points_cached[0, ..., 2], np.full((4, 3), 3000.0))
    assert_array_almost_equal(simplified.points_cached[1, ..., 2], np.full((4, 3), 3100.0))


def test_extract_box_for_well(tmp_path):

    epc = os.path.join(tmp_path, 'tube.epc')

    model = rq.new_model(epc)

    # create a basic block grid with geometry
    grid = grr.RegularGrid(model,
                           extent_kji = (3, 5, 7),
                           origin = (0.0, 0.0, 1000.0),
                           dxyz = (100.0, 100.0, 20.0),
                           title = 'main grid',
                           set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True)
    grid_uuid = grid.uuid

    # create a couple of well trajectories
    cells_visited = [(0, 1, 2), (1, 1, 2), (1, 1, 3), (1, 2, 3), (1, 2, 4), (2, 2, 4)]
    traj_1 = rqw.Trajectory(model,
                            grid = grid,
                            cell_kji0_list = cells_visited,
                            length_uom = 'm',
                            spline_mode = 'linear',
                            well_name = 'well 1')
    traj_2 = rqw.Trajectory(model,
                            grid = grid,
                            cell_kji0_list = cells_visited,
                            length_uom = 'm',
                            spline_mode = 'cube',
                            well_name = 'well 2')
    for traj in (traj_1, traj_2):
        traj.write_hdf5()
        traj.create_xml()
    traj_1_uuid = traj_1.uuid
    traj_2_uuid = traj_2.uuid

    # create a blocked well for one of the trajectories
    assert traj_2.root is not None
    bw = rqw.BlockedWell(model, grid = grid, trajectory = traj_2)
    bw.write_hdf5()
    bw.create_xml()
    bw_uuid = bw.uuid

    # store source model
    model.store_epc()

    # extract box for linear trajectory
    grid_1, box_1 = rqdm.extract_box_for_well(epc_file = epc,
                                              source_grid = grid,
                                              trajectory_uuid = traj_1_uuid,
                                              radius = 120.0,
                                              active_cells_shape = 'tube',
                                              new_grid_title = 'grid 1')

    # check basics of resulting grid
    assert grid_1 is not None
    assert box_1 is not None
    assert tuple(grid_1.extent_kji) == tuple(bx.extent_of_box(box_1))
    expected_box = np.array([(0, 0, 1), (2, 3, 5)], dtype = int)
    assert np.all(box_1 == expected_box)
    #   expected_inactive_1 = np.array(
    #      [[[1, 0, 1, 1, 1], [0, 0, 0, 1, 1], [1, 0, 1, 1, 1], [1, 1, 1, 1, 1]],
    #       [[1, 0, 0, 1, 1], [0, 0, 0, 0, 1], [1, 0, 0, 0, 0], [1, 1, 0, 0, 1]],
    #       [[1, 1, 1, 1, 1], [1, 1, 1, 0, 1], [1, 1, 0, 0, 0], [1, 1, 1, 0, 1]]], dtype = bool)   expected_inactive_1 = np.array(
    expected_inactive_1 = np.array([[[1, 0, 1, 1, 1], [0, 0, 0, 0, 1], [1, 0, 0, 0, 1], [1, 1, 0, 0, 1]],
                                    [[1, 0, 0, 1, 1], [0, 0, 0, 0, 1], [1, 0, 0, 0, 0], [1, 1, 0, 0, 1]],
                                    [[1, 0, 0, 1, 1], [1, 0, 0, 0, 1], [1, 0, 0, 0, 0], [1, 1, 1, 0, 1]]],
                                   dtype = bool)
    assert np.all(grid_1.inactive == expected_inactive_1)

    # extract box for blocked well made from splined trajectory
    grid_2, box_2 = rqdm.extract_box_for_well(epc_file = epc,
                                              source_grid = grid,
                                              blocked_well_uuid = bw_uuid,
                                              radius = 120.0,
                                              active_cells_shape = 'prism',
                                              new_grid_title = 'grid 2')
    assert grid_2 is not None
    assert box_2 is not None
    assert tuple(grid_2.extent_kji) == tuple(bx.extent_of_box(box_2))
    assert np.all(box_2 == expected_box)
    # active cells should be superset of those for linear trajectory tube box
    assert np.count_nonzero(grid_2.inactive) <= np.count_nonzero(grid_1.inactive)
    assert np.all(np.logical_not(grid_2.inactive[np.logical_not(expected_inactive_1)]))
    # check prism shape to inactive cells
    assert np.all(grid_2.inactive == grid_2.inactive[0])


def test_extract_box(tmp_path):
    # create an empty model and add a crs
    epc = os.path.join(tmp_path, 'box_test.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    model.store_epc()
    # create a grid
    grid = grr.RegularGrid(model,
                           crs_uuid = crs.uuid,
                           extent_kji = (5, 10, 12),
                           origin = (2000.0, 3000.0, 1000.0),
                           dxyz = (100.0, 100.0, 20.0),
                           title = 'original grid',
                           as_irregular_grid = True)
    grid.write_hdf5()
    grid.create_xml()
    # introduce a couple of faults
    fault_lines = []
    fl1 = np.array([(2600.0, 2900.0, 0.0), (2600.0, 4100.0, 0.0)])
    fault_lines.append(rql.Polyline(model, set_bool = False, set_coord = fl1, set_crs = crs.uuid, title = 'fault 1'))
    fl2 = np.array([(1900.0, 3500.0, 0.0), (3300.0, 3500.0, 0.0)])
    fault_lines.append(rql.Polyline(model, set_bool = False, set_coord = fl2, set_crs = crs.uuid, title = 'fault 2'))
    for fl in fault_lines:
        fl.write_hdf5()
        fl.create_xml()
    model.store_epc()
    fs_grid = rqdm.add_faults(epc, grid, polylines = fault_lines, new_grid_title = 'small throws')
    # scale the throw on the faults
    fb_grid = rqdm.global_fault_throw_scaling(epc,
                                              source_grid = fs_grid,
                                              scaling_factor = 30.0,
                                              cell_range = 2,
                                              new_grid_title = 'big throws')
    # extract a box
    box = np.array([(1, 2, 3), (3, 8, 9)], dtype = int)
    e_grid = rqdm.extract_box(epc, source_grid = fb_grid, box = box, new_grid_title = 'extracted grid')
    assert e_grid is not None
    # re-open model and check extent of extracted grid
    model = rq.Model(epc)
    grid = grr.Grid(model, uuid = e_grid.uuid)
    assert np.all(grid.extent_kji == (3, 7, 7))


def test_add_grid_points_property(tmp_path):

    epc = os.path.join(tmp_path, 'bland.epc')
    new_epc = os.path.join(tmp_path, 'pointy.epc')

    model = rq.new_model(epc)

    # create a basic block grid with geometry
    extent_kji = (3, 5, 2)
    grid = grr.RegularGrid(model,
                           extent_kji = extent_kji,
                           origin = (2000.0, 3000.0, 1000.0),
                           dxyz = (10.0, 10.0, 20.0),
                           title = 'the grid',
                           set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True, add_cell_length_properties = False)
    grid_uuid = grid.uuid

    # store grid
    model.store_epc()

    # create a points property array
    diagonal = grid.axial_lengths_kji()
    diagonals_extent = tuple(list(extent_kji) + [3])
    diagonal_array = np.empty(diagonals_extent)
    diagonal_array[:] = np.array(diagonal).reshape(1, 1, 1, 3)

    # add to model using derived model function but save as new dataset
    rqdm.add_one_grid_property_array(epc_file = epc,
                                     a = diagonal_array,
                                     property_kind = 'length',
                                     grid_uuid = grid_uuid,
                                     source_info = 'test',
                                     title = 'diagonal vectors',
                                     discrete = False,
                                     uom = grid.xy_units(),
                                     points = True,
                                     extra_metadata = {'test': 'true'},
                                     new_epc_file = new_epc)

    # re-open the original model and check that the points property is not there
    model = rq.Model(epc)
    grid = model.grid()
    pc = grid.property_collection
    assert pc is not None
    assert len(pc.selective_parts_list(points = True)) == 0

    # re-open the new model and load the points property
    model = rq.Model(new_epc)
    grid = model.grid()
    pc = grid.property_collection
    assert pc is not None
    assert len(pc.selective_parts_list(points = True)) == 1
    diag = pc.single_array_ref(points = True)
    assert_array_almost_equal(diag, diagonal_array)


def test_add_edges_per_column_property_array(tmp_path):

    # create a new model with a grid
    epc = os.path.join(tmp_path, 'edges_per_column.epc')
    model = rq.new_model(epc)
    grid = grr.RegularGrid(model, extent_kji = (2, 3, 4))
    grid.write_hdf5()
    grid.create_xml()
    model.store_epc()

    # fabricate an edges per column property
    edge_prop = np.zeros((grid.nj, grid.ni, 2, 2))
    edge_prop[:] = np.linspace(0.1, 0.9, num = edge_prop.size).reshape(edge_prop.shape)

    # add the edges per column property
    prop_uuid = rqdm.add_edges_per_column_property_array(epc,
                                                         edge_prop,
                                                         property_kind = 'multiplier',
                                                         grid_uuid = grid.uuid,
                                                         source_info = 'unit testing',
                                                         title = 'test property on column edges',
                                                         discrete = False,
                                                         uom = 'm3/m3')
    assert prop_uuid is not None

    # re-open the model and inspect the property
    model = rq.Model(epc)
    assert len(model.parts(obj_type = 'ContinuousProperty')) > 0
    edge_property = rqp.Property(model, uuid = prop_uuid)
    assert edge_property is not None
    ep_array = edge_property.array_ref()
    # RESQML holds array with last two dimensions flattened and reordered
    assert ep_array.shape == (grid.nj, grid.ni, 4)
    # restore logical resqpy order and shape
    ep_restored = rqp.reformat_column_edges_from_resqml_format(ep_array)
    assert_array_almost_equal(ep_restored, edge_prop)
    assert edge_property.is_continuous()
    assert not edge_property.is_categorical()
    assert edge_property.indexable_element() == 'edges per column'
    assert edge_property.uom() == 'm3/m3'
    assert edge_property.property_kind() == 'multiplier'
    assert edge_property.facet() is None


def test_add_one_blocked_well_property(example_model_with_well):
    model, well_interp, datum, traj = example_model_with_well
    epc = model.epc_file
    # make a grid positioned in the area of the well
    grid = grr.RegularGrid(model,
                           crs_uuid = model.crs_uuid,
                           origin = (-150.0, -150.0, 1500.0),
                           extent_kji = (5, 3, 3),
                           dxyz = (100.0, 100.0, 50.0),
                           set_points_cached = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True, add_cell_length_properties = False)
    # create a blocked well
    bw = rqw.BlockedWell(model, grid = grid, trajectory = traj)
    bw.write_hdf5()
    bw.create_xml()
    model.store_epc()
    assert bw is not None
    assert bw.cell_count == grid.nk
    # fabricate a blocked well property holding the depths of the centres of penetrated cells
    wb_prop = np.zeros(grid.nk,)
    cells_kji = bw.cell_indices_kji0()
    for i, cell_kji in enumerate(cells_kji):
        wb_prop[i] = grid.centre_point(cell_kji)[2]
    # add the property
    p_uuid = rqdm.add_one_blocked_well_property(epc,
                                                wb_prop,
                                                'depth',
                                                bw.uuid,
                                                source_info = 'unit test',
                                                title = 'DEPTH',
                                                discrete = False,
                                                uom = 'm',
                                                indexable_element = 'cells')
    assert p_uuid is not None
    # re-open the model and check the wellbore property
    model = rq.Model(epc)
    prop = rqp.Property(model, uuid = p_uuid)
    assert prop is not None
    assert bu.matching_uuids(model.supporting_representation_for_part(model.part(uuid = prop.uuid)), bw.uuid)
    assert prop.title == 'DEPTH'
    assert prop.property_kind() == 'depth'
    assert prop.uom() == 'm'
    assert prop.is_continuous()
    assert not prop.is_categorical()
    assert prop.minimum_value() is not None and prop.maximum_value() is not None
    assert prop.minimum_value() > grid.xyz_box(lazy = True)[0, 2]  # minimum z
    assert prop.maximum_value() < grid.xyz_box(lazy = True)[1, 2]  # maximum z


def test_add_wells_from_ascii_file(tmp_path):
    # create an empty model and add a crs
    epc = os.path.join(tmp_path, 'well_model.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    model.store_epc()
    # fabricate some test data as an ascii table
    well_file = os.path.join(tmp_path, 'well_table.txt')
    df = pd.DataFrame(columns = ['WELL', 'MD', 'X', 'Y', 'Z'])
    well_count = 3
    for wi in range(well_count):
        well_name = 'Hole_' + str(wi + 1)
        wdf = pd.DataFrame(columns = ['WELL', 'MD', 'X', 'Y', 'Z'])
        row_count = 3 + wi
        wdf['MD'] = np.linspace(0.0, 1000.0, num = row_count)
        wdf['X'] = np.linspace(100.0 * wi, 100.0 * wi + 10.0 * row_count, num = row_count)
        wdf['Y'] = np.linspace(500.0 * wi, 500.0 * wi - 15.0 * row_count, num = row_count)
        wdf['Z'] = np.linspace(0.0, 1000.0 + 5.0 * row_count, num = row_count)
        wdf['WELL'] = well_name
        df = df.append(wdf)
    df.to_csv(well_file, sep = ' ', index = False)
    # call the derived model function to add the wells
    added = rqdm.add_wells_from_ascii_file(epc,
                                           crs.uuid,
                                           well_file,
                                           space_separated_instead_of_csv = True,
                                           length_uom = 'ft',
                                           md_domain = ['driller', 'logger'][wi % 2],
                                           drilled = True,
                                           z_inc_down = True)
    assert added == well_count
    # re-open the model and check that all expected objects have appeared
    model = rq.Model(epc)
    assert len(model.parts(obj_type = 'WellboreTrajectoryRepresentation')) == well_count
    assert len(model.parts(obj_type = 'MdDatum')) == well_count
    assert len(model.parts(obj_type = 'WellboreInterpretation')) == well_count
    assert len(model.parts(obj_type = 'WellboreFeature')) == well_count
    for wi in range(well_count):
        well_name = 'Hole_' + str(wi + 1)
        traj = rqw.Trajectory(model,
                              uuid = model.uuid(obj_type = 'WellboreTrajectoryRepresentation', title = well_name))
        assert traj is not None
        assert traj.knot_count == 3 + wi


def test_interpolated_grid(tmp_path):
    # create an empty model and add a crs
    epc = os.path.join(tmp_path, 'interpolation.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    # create a pair of grids to act as boundary case geometries
    grid0 = grr.RegularGrid(model,
                            crs_uuid = model.crs_uuid,
                            origin = (0.0, 0.0, 1000.0),
                            extent_kji = (5, 4, 3),
                            dxyz = (100.0, 150.0, 50.0),
                            set_points_cached = True)
    grid0.grid_representation = 'IjkGrid'  # overwrite block grid setting
    grid0.write_hdf5()
    grid0.create_xml(write_geometry = True, add_cell_length_properties = False)
    grid1 = grr.RegularGrid(model,
                            crs_uuid = model.crs_uuid,
                            origin = (15.0, 35.0, 1030.0),
                            extent_kji = (5, 4, 3),
                            dxyz = (97.0, 145.0, 47.0),
                            set_points_cached = True)
    grid1.grid_representation = 'IjkGrid'  # overwrite block grid setting
    grid1.write_hdf5()
    grid1.create_xml(write_geometry = True, add_cell_length_properties = False)
    model.store_epc()
    # interpolate between the two grids
    between_grid_uuids = []
    for f in [0.0, 0.23, 0.5, 1.0]:
        grid = rqdm.interpolated_grid(epc,
                                      grid0,
                                      grid1,
                                      a_to_b_0_to_1 = f,
                                      split_tolerance = 0.01,
                                      inherit_properties = False,
                                      inherit_realization = None,
                                      inherit_all_realizations = False,
                                      new_grid_title = 'between_' + str(f))
        assert grid is not None
        between_grid_uuids.append(grid.uuid)
    # re-open model and check interpolated grid geometries
    model = rq.Model(epc)
    for i, g_uuid in enumerate(between_grid_uuids):
        grid = grr.Grid(model, uuid = g_uuid)
        assert grid is not None
        grid.cache_all_geometry_arrays()
        assert hasattr(grid, 'points_cached') and grid.points_cached is not None
        assert np.all(grid.points_cached >= grid0.points_cached)
        assert np.all(grid.points_cached <= grid1.points_cached)
        if i == 0:
            assert_array_almost_equal(grid.points_cached, grid0.points_cached)
        elif i == len(between_grid_uuids) - 1:
            assert_array_almost_equal(grid.points_cached, grid1.points_cached)
        else:
            assert not np.any(np.isclose(grid.points_cached, grid0.points_cached))
            assert not np.any(np.isclose(grid.points_cached, grid1.points_cached))


def test_interpolated_faulted_grid(tmp_path):
    # create an empty model and add a crs
    epc = os.path.join(tmp_path, 'faulted_interpolation.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    # create a regular grid
    grid0 = grr.RegularGrid(model,
                            crs_uuid = model.crs_uuid,
                            origin = (0.0, 0.0, 1000.0),
                            extent_kji = (5, 4, 3),
                            dxyz = (100.0, 150.0, 50.0),
                            as_irregular_grid = True)
    grid0.grid_representation = 'IjkGrid'  # overwrite block grid setting
    grid0.write_hdf5()
    grid0.create_xml(write_geometry = True, add_cell_length_properties = False)
    model.store_epc()
    # prepare fault data and add faults to grid with two different throw sets
    pl_dict = {}
    pl_dict['f1'] = [(2, 0), (2, 1), (1, 1), (1, 2), (1, 3)]
    pl_dict['f2'] = [(0, 2), (1, 2), (2, 2), (3, 2), (4, 2)]
    lr_dict = {}
    lr_dict['f1'] = (23.0, -23.0)
    lr_dict['f2'] = (-37.3, 14.8)
    grid_a = rqdm.add_faults(epc,
                             source_grid = grid0,
                             full_pillar_list_dict = pl_dict,
                             left_right_throw_dict = lr_dict,
                             new_grid_title = 'grid a')
    lr_dict = {}
    lr_dict['f1'] = (-23.0, 23.0)
    lr_dict['f2'] = (17.2, -22.9)
    grid_b = rqdm.add_faults(epc,
                             source_grid = grid0,
                             full_pillar_list_dict = pl_dict,
                             left_right_throw_dict = lr_dict,
                             new_grid_title = 'grid b')
    # interpolate between the two grids
    between_grid_uuids = []
    for f in [0.0, 0.23, 0.5, 1.0]:
        grid = rqdm.interpolated_grid(epc,
                                      grid_a,
                                      grid_b,
                                      a_to_b_0_to_1 = f,
                                      split_tolerance = 0.01,
                                      inherit_properties = False,
                                      inherit_realization = None,
                                      inherit_all_realizations = False,
                                      new_grid_title = 'between_' + str(f))
        assert grid is not None
        between_grid_uuids.append(grid.uuid)
    # re-open model and check end point interpolated grid geometries
    model = rq.Model(epc)
    for i, g_uuid in enumerate(between_grid_uuids):
        grid = grr.Grid(model, uuid = g_uuid)
        assert grid is not None
        grid.cache_all_geometry_arrays()
        assert hasattr(grid, 'points_cached') and grid.points_cached is not None
        if i == 0:
            assert_array_almost_equal(grid.points_cached, grid_a.points_cached)
        elif i == len(between_grid_uuids) - 1:
            assert_array_almost_equal(grid.points_cached, grid_b.points_cached)


def test_interpolated_grid_using_cp(tmp_path):
    # create an empty model and add a crs
    epc = os.path.join(tmp_path, 'cp_interpolation.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    # create a regular grid
    grid0 = grr.RegularGrid(model,
                            crs_uuid = model.crs_uuid,
                            origin = (0.0, 0.0, 1000.0),
                            extent_kji = (5, 4, 3),
                            dxyz = (100.0, 150.0, 50.0),
                            as_irregular_grid = True)
    grid0.grid_representation = 'IjkGrid'  # overwrite block grid setting
    grid0.write_hdf5()
    grid0.create_xml(write_geometry = True, add_cell_length_properties = False)
    model.store_epc()
    # prepare fault data and add faults to a copy of the grid
    pl_dict = {}
    pl_dict['f1'] = [(2, 0), (2, 1), (1, 1), (1, 2), (1, 3)]
    pl_dict['f2'] = [(0, 2), (1, 2), (2, 2), (3, 2), (4, 2)]
    lr_dict = {}
    lr_dict['f1'] = (23.0, -23.0)
    lr_dict['f2'] = (-37.3, 14.8)
    grid_a = rqdm.add_faults(epc,
                             source_grid = grid0,
                             full_pillar_list_dict = pl_dict,
                             left_right_throw_dict = lr_dict,
                             new_grid_title = 'grid a')
    # interpolate between the two grids
    between_grid_uuids = []
    for f in [0.0, 0.23, 0.5, 1.0]:
        grid = rqdm.interpolated_grid(epc,
                                      grid0,
                                      grid_a,
                                      a_to_b_0_to_1 = f,
                                      split_tolerance = 0.01,
                                      inherit_properties = False,
                                      inherit_realization = None,
                                      inherit_all_realizations = False,
                                      new_grid_title = 'between_' + str(f))
        assert grid is not None
        between_grid_uuids.append(grid.uuid)
    # re-open model and check end point interpolated grid geometries
    model = rq.Model(epc)
    for i, g_uuid in enumerate(between_grid_uuids):
        grid = grr.Grid(model, uuid = g_uuid)
        assert grid is not None
        grid.cache_all_geometry_arrays()
        assert hasattr(grid, 'points_cached') and grid.points_cached is not None
        if i == 0:
            assert_array_almost_equal(grid.corner_points(), grid0.corner_points(), decimal = 3)
        elif i == len(between_grid_uuids) - 1:
            assert_array_almost_equal(grid.corner_points(), grid_a.corner_points(), decimal = 3)


def test_refined_and_coarsened_grid(tmp_path):

    # create a model and a coarse grid
    epc = os.path.join(tmp_path, 'refinement.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()

    # create a coarse grid
    c_dxyz = (100.0, 150.0, 50.0)
    c_grid = grr.RegularGrid(model,
                             crs_uuid = model.crs_uuid,
                             extent_kji = (5, 3, 4),
                             dxyz = c_dxyz,
                             as_irregular_grid = True)
    c_grid.write_hdf5()
    c_grid.create_xml(write_geometry = True, add_cell_length_properties = True, expand_const_arrays = True)
    model.store_epc()

    # set up a coarse to fine mapping
    fine_extent = np.array(c_grid.extent_kji)
    fine_extent[0] *= 2
    fine_extent[1] *= 4
    fine_extent[2] *= 3
    fc = rqfc.FineCoarse(fine_extent, c_grid.extent_kji)
    fc.set_all_ratios_constant()

    # make the refinement
    f_grid = rqdm.refined_grid(epc,
                               source_grid = None,
                               fine_coarse = fc,
                               inherit_properties = True,
                               inherit_realization = None,
                               inherit_all_realizations = False,
                               source_grid_uuid = c_grid.uuid,
                               set_parent_window = None,
                               infill_missing_geometry = True,
                               new_grid_title = 'fine grid')
    assert f_grid is not None

    # re-open model and check some things about the refined grid
    model = rq.Model(epc)
    f_grid = grr.Grid(model, uuid = f_grid.uuid)
    assert tuple(f_grid.extent_kji) == tuple(fine_extent)
    f_grid.cache_all_geometry_arrays()
    assert_array_almost_equal(c_grid.xyz_box(lazy = False), f_grid.xyz_box(lazy = False))
    assert not f_grid.has_split_coordinate_lines
    assert f_grid.points_cached is not None
    assert f_grid.points_cached.shape == (c_grid.nk * 2 + 1, c_grid.nj * 4 + 1, c_grid.ni * 3 + 1, 3)
    # check that refined grid geometry is monotonic in each of the three axes
    p = f_grid.points_ref(masked = False)
    assert np.all(p[:-1, :, :, 2] < p[1:, :, :, 2])
    assert np.all(p[:, :-1, :, 1] < p[:, 1:, :, 1])
    assert np.all(p[:, :, :-1, 0] < p[:, :, 1:, 0])

    # check property inheritance of cell lengths
    pc = f_grid.extract_property_collection()
    assert pc is not None and pc.number_of_parts() >= 3
    lpc = rqp.selective_version_of_collection(pc, property_kind = 'cell length')
    assert lpc.number_of_parts() == 3
    for axis in range(3):
        length_array = lpc.single_array_ref(facet_type = 'direction', facet = 'KJI'[axis])
        assert length_array is not None
        assert np.allclose(length_array, c_dxyz[2 - axis] / (2, 4, 3)[axis])

    # make a (re-)coarsened version of the fine grid
    cfc_grid = rqdm.coarsened_grid(epc,
                                   f_grid,
                                   fc,
                                   inherit_properties = True,
                                   set_parent_window = None,
                                   infill_missing_geometry = True,
                                   new_grid_title = 're-coarsened grid')
    assert cfc_grid is not None

    # re-open the model and re-load the new grid
    model = rq.Model(epc)
    assert len(model.parts(obj_type = 'IjkGridRepresentation')) == 3
    cfc_grid = grr.Grid(model, uuid = cfc_grid.uuid)

    # compare the re-coarsened grid with the original coarse grid
    assert tuple(cfc_grid.extent_kji) == tuple(c_grid.extent_kji)
    assert not cfc_grid.has_split_coordinate_lines
    cfc_grid.cache_all_geometry_arrays()
    assert_array_almost_equal(c_grid.points_cached, cfc_grid.points_cached)

    # see how the cell length properties have fared
    cfc_pc = cfc_grid.extract_property_collection()
    assert cfc_grid is not None
    assert cfc_pc.number_of_parts() >= 3
    cfc_lpc = rqp.selective_version_of_collection(cfc_pc, property_kind = 'cell length')
    assert cfc_lpc.number_of_parts() == 3
    for axis in range(3):
        length_array = cfc_lpc.single_array_ref(facet_type = 'direction', facet = 'KJI'[axis])
        assert length_array is not None
        assert np.allclose(length_array, c_dxyz[2 - axis])


def test_add_faults(tmp_path):

    def write_poly(filename, a, mode = 'w'):
        nines = 999.0
        with open(filename, mode = mode) as fp:
            for row in range(len(a)):
                fp.write(f'{a[row, 0]:8.3f} {a[row, 1]:8.3f} {a[row, 2]:8.3f}\n')
            fp.write(f'{nines:8.3f} {nines:8.3f} {nines:8.3f}\n')

    def make_poly(model, a, title, crs):
        return [rql.Polyline(model, set_bool = False, set_coord = a, set_crs = crs.uuid, title = title)]

    epc = os.path.join(tmp_path, 'tic_tac_toe.epc')

    for test_mode in ['file', 'polyline']:

        model = rq.new_model(epc)
        grid = grr.RegularGrid(model, extent_kji = (1, 3, 3), set_points_cached = True)
        grid.write_hdf5()
        grid.create_xml(write_geometry = True)
        crs = rqc.Crs(model, uuid = grid.crs_uuid)
        model.store_epc()

        # single straight fault
        a = np.array([[-0.2, 2.0, -0.1], [3.2, 2.0, -0.1]])
        f = os.path.join(tmp_path, 'ttt_f1.dat')
        if test_mode == 'file':
            write_poly(f, a)
            lines_file_list = [f]
            polylines = None
        else:
            lines_file_list = None
            polylines = make_poly(model, a, 'ttt_f1', crs)
        g = rqdm.add_faults(epc,
                            source_grid = None,
                            polylines = polylines,
                            lines_file_list = lines_file_list,
                            inherit_properties = False,
                            new_grid_title = 'ttt_f1 straight')

        # single zig-zag fault
        a = np.array([[-0.2, 1.0, -0.1], [1.0, 1.0, -0.1], [1.0, 2.0, -0.1], [3.2, 2.0, -0.1]])
        f = os.path.join(tmp_path, 'ttt_f2.dat')
        if test_mode == 'file':
            write_poly(f, a)
            lines_file_list = [f]
            polylines = None
        else:
            lines_file_list = None
            polylines = make_poly(model, a, 'ttt_f2', crs)
        g = rqdm.add_faults(epc,
                            source_grid = None,
                            polylines = polylines,
                            lines_file_list = lines_file_list,
                            inherit_properties = True,
                            new_grid_title = 'ttt_f2 zig_zag')

        # single zig-zag-zig fault
        a = np.array([[-0.2, 1.0, -0.1], [1.0, 1.0, -0.1], [1.0, 2.0, -0.1], [2.0, 2.0, -0.1], [2.0, 1.0, -0.1],
                      [3.2, 1.0, -0.1]])
        f = os.path.join(tmp_path, 'ttt_f3.dat')
        if test_mode == 'file':
            write_poly(f, a)
            lines_file_list = [f]
            polylines = None
        else:
            lines_file_list = None
            polylines = make_poly(model, a, 'ttt_f3', crs)
        g = rqdm.add_faults(epc,
                            source_grid = None,
                            polylines = polylines,
                            lines_file_list = lines_file_list,
                            inherit_properties = True,
                            new_grid_title = 'ttt_f3 zig_zag_zig')

        # horst block
        a = np.array([[-0.2, 1.0, -0.1], [3.2, 1.0, -0.1]])
        b = np.array([[3.2, 2.0, -0.1], [-0.2, 2.0, -0.1]])
        fa = os.path.join(tmp_path, 'ttt_f4a.dat')
        fb = os.path.join(tmp_path, 'ttt_f4b.dat')
        if test_mode == 'file':
            write_poly(fa, a)
            write_poly(fb, b)
            lines_file_list = [fa, fb]
            polylines = None
        else:
            lines_file_list = None
            polylines = make_poly(model, a, 'ttt_f4a', crs) + make_poly(model, b, 'ttt_f4b', crs)
        g = rqdm.add_faults(epc,
                            source_grid = None,
                            polylines = polylines,
                            lines_file_list = lines_file_list,
                            inherit_properties = True,
                            new_grid_title = 'ttt_f4 horst')

        # asymmetrical horst block
        lr_throw_dict = {'ttt_f4a': (0.0, -0.3), 'ttt_f4b': (0.0, -0.6)}
        g = rqdm.add_faults(epc,
                            source_grid = None,
                            polylines = polylines,
                            lines_file_list = lines_file_list,
                            left_right_throw_dict = lr_throw_dict,
                            inherit_properties = True,
                            new_grid_title = 'ttt_f5 horst')
        assert g is not None

        # scaled version of asymmetrical horst block
        model = rq.Model(epc)
        grid = model.grid(title = 'ttt_f5 horst')
        assert grid is not None
        gcs_uuids = model.uuids(obj_type = 'GridConnectionSetRepresentation', related_uuid = grid.uuid)
        assert gcs_uuids
        scaling_dict = {'ttt_f4a': 3.0, 'ttt_f4b': 1.7}
        for i, gcs_uuid in enumerate(gcs_uuids):
            gcs = rqf.GridConnectionSet(model, uuid = gcs_uuid)
            assert gcs is not None
            assert bu.matching_uuids(gcs.uuid, gcs_uuid)
            rqdm.fault_throw_scaling(epc,
                                     source_grid = grid,
                                     scaling_factor = None,
                                     connection_set = gcs,
                                     scaling_dict = scaling_dict,
                                     ref_k0 = 0,
                                     ref_k_faces = 'top',
                                     cell_range = 0,
                                     offset_decay = 0.5,
                                     store_displacement = False,
                                     inherit_properties = True,
                                     inherit_realization = None,
                                     inherit_all_realizations = False,
                                     new_grid_title = f'ttt_f6 scaled {i+1}',
                                     new_epc_file = None)
            model = rq.Model(epc)
            grid = model.grid(title = f'ttt_f6 scaled {i+1}')
            assert grid is not None

        # two intersecting straight faults
        a = np.array([[-0.2, 2.0, -0.1], [3.2, 2.0, -0.1]])
        b = np.array([[1.0, -0.2, -0.1], [1.0, 3.2, -0.1]])
        f = os.path.join(tmp_path, 'ttt_f7.dat')
        write_poly(f, a)
        write_poly(f, b, mode = 'a')
        if test_mode == 'file':
            write_poly(f, a)
            write_poly(f, b, mode = 'a')
            lines_file_list = [f]
            polylines = None
        else:
            lines_file_list = None
            polylines = make_poly(model, a, 'ttt_f7_1', crs) + make_poly(model, b, 'ttt_f7_2', crs)
        g = rqdm.add_faults(epc,
                            source_grid = None,
                            polylines = polylines,
                            lines_file_list = lines_file_list,
                            inherit_properties = True,
                            new_grid_title = 'ttt_f7')

        # re-open and check a few things
        model = rq.Model(epc)
        assert len(model.titles(obj_type = 'IjkGridRepresentation')) == 8
        g1 = model.grid(title = 'ttt_f7')
        assert g1.split_pillars_count == 5
        cpm = g1.create_column_pillar_mapping()
        assert cpm.shape == (3, 3, 2, 2)
        extras = (cpm >= 16)
        assert np.count_nonzero(extras) == 7
        assert np.all(np.sort(np.unique(cpm)) == np.arange(21))


def test_add_faults_and_scaling(tmp_path):

    # create a model with crs
    epc = os.path.join(tmp_path, 'fault_test.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()

    # create a grid
    dxyz = (100.0, 150.0, 20.0)
    grid = grr.RegularGrid(model,
                           crs_uuid = model.crs_uuid,
                           extent_kji = (3, 5, 5),
                           dxyz = dxyz,
                           as_irregular_grid = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True, add_cell_length_properties = True, expand_const_arrays = True)
    model.store_epc()

    # prepare some polylines to define faults
    pla = rql.Polyline(model,
                       set_bool = False,
                       set_crs = crs.uuid,
                       title = 'line_a',
                       set_coord = np.array([(50.0, -10.0, 0.0), (450.0, 760.0, 0.0)]))
    plb = rql.Polyline(model,
                       set_bool = False,
                       set_crs = crs.uuid,
                       title = 'line_b',
                       set_coord = np.array([(-10.0, 530.0, 0.0), (510.0, 330.0, 0.0)]))
    for pl in (pla, plb):
        pl.write_hdf5()
        pl.create_xml()
    model.store_epc()

    # add faults with default throws
    f1a_grid = rqdm.add_faults(epc, source_grid = grid, polylines = [pla, plb], new_grid_title = 'faulted by polylines')
    f1a_grid_uuid = f1a_grid.uuid

    # prepare dictionaries to define more faults and their throws
    pillar_list_dict = {}
    pillar_list_dict['step_fault'] = [(1, 0), (1, 1), (2, 1), (2, 2), (2, 3), (2, 4), (1, 4), (1, 5)]
    pillar_list_dict['north_south_fault'] = [(0, 3), (4, 3)]
    throw_dict = {}
    throw_dict['step_fault'] = (7.0, -7.0)
    throw_dict['north_south_fault'] = (-3.3, 4.8)

    # add faults with explicit throws
    f2a_grid = rqdm.add_faults(epc,
                               source_grid = grid,
                               full_pillar_list_dict = pillar_list_dict,
                               left_right_throw_dict = throw_dict,
                               new_grid_title = 'faulted by dictionaries')
    f2a_grid_uuid = f2a_grid.uuid

    # scale faults globally
    f1b_grid = rqdm.global_fault_throw_scaling(epc,
                                               source_grid = f1a_grid,
                                               scaling_factor = 10.0,
                                               cell_range = 2,
                                               new_grid_title = 'globally scaled faults')
    f1b_grid_uuid = f1b_grid.uuid

    # re-open the model to identify a grid connection set
    model = rq.Model(epc)
    gcs_uuid = model.uuid(obj_type = 'GridConnectionSetRepresentation', related_uuid = f2a_grid_uuid)
    assert gcs_uuid is not None
    gcs = rqf.GridConnectionSet(model, uuid = gcs_uuid)
    assert gcs is not None

    # create a scaling dictionary
    scaling_dict = {'step_fault': 1.5, 'north_south_fault': 3.0}

    # scale faults
    f2b_grid = rqdm.fault_throw_scaling(epc,
                                        source_grid = f2a_grid,
                                        connection_set = gcs,
                                        scaling_dict = scaling_dict,
                                        cell_range = 1,
                                        new_grid_title = 'dictionary scaled faults')
    f2b_grid_uuid = f2b_grid.uuid

    # re-open model and check grids
    model = rq.Model(epc)
    assert len(model.uuids(obj_type = 'IjkGridRepresentation')) == 5

    g = grr.Grid(model, uuid = f1a_grid_uuid)
    gbox = g.xyz_box(lazy = False)
    assert_array_almost_equal(gbox, np.array([(0.0, 0.0, -1.0), (500.0, 750.0, 61.0)]))

    g = grr.Grid(model, uuid = f2a_grid_uuid)
    gbox = g.xyz_box(lazy = False)
    # assert_array_almost_equal(gbox, np.array([(0.0, 0.0, -10.3), (500.0, 750.0, 71.8)]))
    assert_array_almost_equal(gbox, np.array([(0.0, 0.0, -7.0), (500.0, 750.0, 67.0)]))

    g = grr.Grid(model, uuid = f1b_grid_uuid)
    gbox = g.xyz_box(lazy = False)
    # assert_array_almost_equal(gbox, np.array([(0.0, 0.0, -5.0), (500.0, 750.0, 65.0)]))
    assert_array_almost_equal(gbox, np.array([(0.0, 0.0, -7.25), (500.0, 750.0, 67.75)]))

    g = grr.Grid(model, uuid = f2b_grid_uuid)
    gbox = g.xyz_box(lazy = False)
    assert_array_almost_equal(gbox, np.array([(0.0, 0.0, -10.5), (500.0, 750.0, 70.5)]))


def test_drape_to_surface(tmp_path):

    # create a model and a regular grid
    epc = os.path.join(tmp_path, 'drape_test.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    dxyz = (180.0, -250.0, 25.0)
    grid = grr.RegularGrid(model,
                           crs_uuid = model.crs_uuid,
                           extent_kji = (2, 5, 6),
                           dxyz = dxyz,
                           as_irregular_grid = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True, add_cell_length_properties = False, expand_const_arrays = True)
    model.store_epc()

    # make a wavey surface covering the area of the grid with a boundary buffer
    model = rq.Model(epc)
    source_grid = model.grid()
    xyz_box = source_grid.xyz_box(lazy = False, local = True)
    surf = rqs.Surface(model)
    nx = ny = 100
    mesh_xyz = np.empty((100, 100, 3))
    xy_range = np.max(xyz_box[1, 0:2] - xyz_box[0, 0:2]) * 1.2
    xy_centre = np.mean(xyz_box[:, 0:2], axis = 0)
    origin_xy = xy_centre - xy_range * 0.5
    dx = xy_range / nx
    dy = xy_range / ny
    points_z = source_grid.points_ref()[..., 2]
    thickness = np.nanmean(points_z[-1, ...] - points_z[0, ...])
    x_wavelength = xy_range / 20.0
    x_half_amplitude = thickness / 10.0
    y_wavelength = xy_range / 10.0
    y_half_amplitude = thickness / 4.0
    mesh_xyz[:, :, 2] = xyz_box[0, 2]  # set z initially to flat plane at depth of shallowest point of grid
    for i in range(nx):
        x = dx * i
        mesh_xyz[i, :, 0] = origin_xy[0] + dx * i  # x
        mesh_xyz[i, :, 2] += x_half_amplitude * maths.sin(
            x * 2.0 * maths.pi / x_wavelength)  # add a depth wave in x direction
    for j in range(ny):
        y = dy * j
        mesh_xyz[:, j, 1] = origin_xy[1] + dy * j  # y
        mesh_xyz[:, j, 2] += y_half_amplitude * maths.sin(
            y * 2.0 * maths.pi / y_wavelength)  # add a depth wave in y direction
    surf.set_from_irregular_mesh(mesh_xyz)

    # drape the grid to the surface
    rqdm.drape_to_surface(epc,
                          source_grid,
                          surf,
                          ref_k0 = 0,
                          ref_k_faces = 'top',
                          store_displacement = True,
                          new_grid_title = 'draped')

    # reopen the model
    model = rq.Model(epc)
    draped = model.grid(title = 'draped')
    assert draped is not None


def test_zonal_grid(tmp_path):

    # create a model and a regular grid
    epc = os.path.join(tmp_path, 'zonal_test.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    dxyz = (200.0, -270.0, 12.0)
    grid = grr.RegularGrid(model,
                           crs_uuid = model.crs_uuid,
                           extent_kji = (9, 2, 3),
                           dxyz = dxyz,
                           as_irregular_grid = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True, add_cell_length_properties = False)
    model.store_epc()

    # create a zonal version of the grid
    zone_ranges = [(0, 1, 0), (2, 4, 1), (5, 8, 2), (9, 9, 3)]
    rqdm.zonal_grid(epc, zone_layer_range_list = zone_ranges, new_grid_title = 'four zone grid')

    # create a single layer version of the grid
    rqdm.single_layer_grid(epc, source_grid = grid, k0_min = 1, k0_max = 7, new_grid_title = 'single layer grid')

    # re-open the model and take a look at the zonal grids
    model = rq.Model(epc)
    z_grid = model.grid(title = 'four zone grid')
    assert z_grid.nk == 4
    o_grid = model.grid(title = 'single layer grid')
    assert o_grid.nk == 1
    # check z range of single layer grid
    o_box = o_grid.xyz_box()
    assert maths.isclose(o_box[1, 2] - o_box[0, 2], 7.0 * 12.0)


def test_unsplit_grid(tmp_path):

    # create a model and a regular grid
    epc = os.path.join(tmp_path, 'unsplit_test.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    dxyz = (100.0, 150.0, 30.0)
    grid = grr.RegularGrid(model,
                           crs_uuid = model.crs_uuid,
                           extent_kji = (2, 3, 5),
                           dxyz = dxyz,
                           as_irregular_grid = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True, add_cell_length_properties = False)
    model.store_epc()

    # add a fault
    # prepare dictionaries to define more faults and their throws
    pillar_list_dict = {}
    pillar_list_dict['fault'] = [(0, 2), (1, 2), (1, 3), (2, 3)]
    throw_dict = {}
    throw_dict['fault'] = (7.0, -7.0)
    f2a_grid = rqdm.add_faults(epc,
                               source_grid = grid,
                               full_pillar_list_dict = pillar_list_dict,
                               left_right_throw_dict = throw_dict,
                               new_grid_title = 'faulted by dictionaries')
    f2a_grid_uuid = f2a_grid.uuid

    # heal faults
    healed_grid = rqdm.unsplit_grid(epc, source_grid = f2a_grid, new_grid_title = 'healed grid')
    assert healed_grid is not None
    assert not healed_grid.has_split_coordinate_lines


def test_tilted_grid(tmp_path):

    # create a model and a regular grid
    epc = os.path.join(tmp_path, 'tilted_test.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    dxyz = (50.0, 50.0, 10.0)
    grid = grr.RegularGrid(model,
                           crs_uuid = model.crs_uuid,
                           extent_kji = (2, 2, 2),
                           origin = (200.0, 200.0, 1000.0),
                           dxyz = dxyz,
                           as_irregular_grid = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True, add_cell_length_properties = False)
    model.store_epc()

    # tilt by 45 degrees
    t_grid = rqdm.tilted_grid(epc,
                              pivot_xyz = (250.0, 250.0, 1000.0),
                              azimuth = 90.0,
                              dip = 45.0,
                              new_grid_title = 'tilted')
    assert t_grid is not None

    # check xyz box of tilted grid
    root_two = maths.sqrt(2.0)
    expected_box = np.array([(250.0 - 70.0 / root_two, 200.0, 1000.0 - 50.0 / root_two),
                             (250.0 + 50.0 / root_two, 300.0, 1000.0 + 70.0 / root_two)])
    assert_array_almost_equal(t_grid.xyz_box(lazy = False), expected_box)


def test_local_depth_adjustment(tmp_path):

    # create a model and a regular grid
    epc = os.path.join(tmp_path, 'depth_adjust_test.epc')
    model = rq.new_model(epc)
    crs = rqc.Crs(model)
    crs.create_xml()
    dxyz = (50.0, 50.0, 10.0)
    grid = grr.RegularGrid(model,
                           crs_uuid = model.crs_uuid,
                           extent_kji = (2, 20, 20),
                           origin = (500.0, 700.0, 2000.0),
                           dxyz = dxyz,
                           as_irregular_grid = True)
    grid.write_hdf5()
    grid.create_xml(write_geometry = True, add_cell_length_properties = False)
    model.store_epc()

    a_grid = rqdm.local_depth_adjustment(epc, grid, 700.0, 1000.0, 170.0, -7.0, False)
    assert a_grid is not None
    p = a_grid.points_ref()
    assert maths.isclose(np.min(p[..., 2]), 2000.0 - 7.0)


def test_gather_ensemble(tmp_path):
    # create three models, each with a regular grid and a grid property
    epc_list = []
    for m in range(3):
        epc = os.path.join(tmp_path, f'grid_{m}.epc')
        model = rq.new_model(epc)
        crs = rqc.Crs(model)
        crs.create_xml()
        dxyz = (50.0, 50.0, 10.0)
        extent_kji = (3, 20, 20)
        grid = grr.RegularGrid(model,
                               crs_uuid = model.crs_uuid,
                               extent_kji = extent_kji,
                               origin = (500.0, 700.0, 1000.0),
                               dxyz = dxyz,
                               as_irregular_grid = True)
        grid.write_hdf5()
        grid.create_xml(write_geometry = True, add_cell_length_properties = False)
        model.store_epc()
        ntg = np.random.random(extent_kji)
        rqdm.add_one_grid_property_array(epc,
                                         ntg,
                                         property_kind = 'net to gross ratio',
                                         grid_uuid = grid.uuid,
                                         title = 'NETGRS',
                                         uom = 'm3/m3',
                                         indexable_element = 'cells')
        epc_list.append(epc)
    # gather ensemble
    combined_epc = os.path.join(tmp_path, 'combo.epc')
    rqdm.gather_ensemble(epc_list, combined_epc)
    # open combined model and check realisations
    model = rq.Model(combined_epc)
    grid = model.grid()
    assert grid is not None
    pc = grid.property_collection
    assert pc.has_multiple_realizations()
    assert pc.realization_list(sort_list = True) == [0, 1, 2]
    ntg_pc = rqp.selective_version_of_collection(pc, property_kind = 'net to gross ratio')
    assert ntg_pc.number_of_parts() == 3
    ntg3 = ntg_pc.realizations_array_ref()
    assert ntg3.shape == (3, 3, 20, 20)
    assert np.all(ntg3 >= 0.0) and np.all(ntg3 <= 1.0)



import os
import numpy as np
from numpy.testing import assert_array_almost_equal
import pandas as pd
import pytest

from resqpy.model import Model
import resqpy.well


def test_DeviationSurvey(example_model_with_well, tmp_path):
    # Test that all attributes are correctly saved and loaded from disk

    # --------- Arrange ----------
    # Create a Deviation Survey object in memory

    # Load example model from a fixture
    model, well_interp, datum, traj = example_model_with_well
    epc_path = model.epc_file

    # Create 3 copies of the survey, using different initialisers
    data = dict(
        title = 'Majestic Umlaut ö',
        originator = 'Thor, god of sparkles',
        md_uom = 'ft',
        angle_uom = 'rad',
        is_final = True,
    )
    array_data = dict(
        measured_depths = np.array([1, 2, 3], dtype = float) + 1000.0,
        azimuths = np.array([4, 5, 6], dtype = float),
        inclinations = np.array([1, 2, 3], dtype = float),
        first_station = np.array([0, -1, 999], dtype = float),
    )

    survey = resqpy.well.DeviationSurvey(
        parent_model = model,
        represented_interp = well_interp,
        md_datum = datum,
        **data,
        **array_data,
    )
    survey_uuid = survey.uuid

    df = pd.DataFrame(columns = ['MD', 'AZIM_GN', 'INCL', 'X', 'Y', 'Z'])
    for col, key in zip(('MD', 'AZIM_GN', 'INCL'), ('measured_depths', 'azimuths', 'inclinations')):
        df[col] = array_data[key]
    for axis, col in enumerate(('X', 'Y', 'Z')):
        df[col] = np.NaN
        df.loc[0, col] = array_data['first_station'][axis]

    survey_b = resqpy.well.DeviationSurvey.from_data_frame(parent_model = model,
                                                           data_frame = df,
                                                           md_datum = datum,
                                                           md_uom = data['md_uom'],
                                                           angle_uom = data['angle_uom'])
    survey_b_uuid = survey_b.uuid

    csv_file = os.path.join(tmp_path, 'survey_c.csv')
    df.to_csv(csv_file)

    survey_c = resqpy.well.DeviationSurvey.from_ascii_file(parent_model = model,
                                                           deviation_survey_file = csv_file,
                                                           md_datum = datum,
                                                           md_uom = data['md_uom'],
                                                           angle_uom = data['angle_uom'])
    survey_c_uuid = survey_c.uuid

    # ----------- Act ---------

    # Save to disk
    for s in (survey, survey_b, survey_c):
        s.write_hdf5()
        s.create_xml()
    model.store_epc()
    model.h5_release()

    # Clear memory
    del model, well_interp, datum, traj, survey, survey_b, survey_c

    # Reload from disk
    model2 = Model(epc_file = epc_path)
    survey2 = resqpy.well.DeviationSurvey(model2, uuid = survey_uuid)
    survey_b2 = resqpy.well.DeviationSurvey(model2, uuid = survey_b_uuid)
    survey_c2 = resqpy.well.DeviationSurvey(model2, uuid = survey_c_uuid)

    # --------- Assert --------------
    # Check all attributes were loaded from disk correctly

    for key, expected_value in data.items():
        assert getattr(survey2, key) == expected_value, f"Error for {key}"
        if 'uom' in key:
            for s in (survey_b2, survey_c2):
                assert getattr(s, key) == expected_value, f"Error for {key}"

    for s in (survey2, survey_b2, survey_c2):
        for key, expected_value in array_data.items():
            assert_array_almost_equal(getattr(s, key), expected_value, err_msg = f"Error for {key}")
        assert s.station_count == len(array_data['azimuths'])


def test_get_md_datum(example_model_with_well):
    # Test that a ValueError is raised if a valid Md Datum root node has not been provided and cannot be created.

    # --------- Arrange ----------
    # Create a Deviation Survey object in memory

    # Load example model from a fixture
    model, well_interp, datum, traj = example_model_with_well
    epc_path = model.epc_file

    # Create a survey
    data = dict(
        title = 'Majestic Umlaut ö',
        originator = 'Thor, god of sparkles',
        md_uom = 'ft',
        angle_uom = 'rad',
        is_final = True,
    )
    array_data = dict(
        measured_depths = np.array([1, 2, 3], dtype = float) + 1000.0,
        azimuths = np.array([4, 5, 6], dtype = float),
        inclinations = np.array([1, 2, 3], dtype = float),
        first_station = np.array([0, -1, 999], dtype = float),
    )

    survey = resqpy.well.DeviationSurvey(
        parent_model = model,
        represented_interp = well_interp,
        md_datum = None,
        **data,
        **array_data,
    )

    # ----------- Act ---------
    survey.write_hdf5()
    with pytest.raises(ValueError) as excinfo:
        survey._DeviationSurvey__get_md_datum_root(md_datum_root = None, md_datum_xyz = None)

    # -------- Assert ---------
    assert "Must provide a MD Datum for the DeviationSurvey" in str(excinfo.value)



"""Class for RESQML Rock Fluid Unit Feature organizational objects."""

from ._utils import (equivalent_extra_metadata, alias_for_attribute, extract_has_occurred_during,
                     equivalent_chrono_pairs, create_xml_has_occurred_during)

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns
from .boundary_feature import BoundaryFeature


class RockFluidUnitFeature(BaseResqpy):
    """Class for RESQML Rock Fluid Unit Feature organizational objects."""

    resqml_type = "RockFluidUnitFeature"
    feature_name = alias_for_attribute("title")
    valid_phases = ('aquifer', 'gas cap', 'oil column', 'seal')

    def __init__(self,
                 parent_model,
                 root_node = None,
                 uuid = None,
                 phase = None,
                 feature_name = None,
                 top_boundary_feature = None,
                 base_boundary_feature = None,
                 extra_metadata = None):
        """Initialises a rock fluid unit feature organisational object."""

        self.phase = phase
        self.top_boundary_feature = top_boundary_feature
        self.base_boundary_feature = base_boundary_feature

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = feature_name,
                         extra_metadata = extra_metadata,
                         root_node = root_node)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this feature is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, RockFluidUnitFeature):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.feature_name != other.feature_name or self.phase != other.phase:
            return False
        if self.top_boundary_feature is not None:
            if not self.top_boundary_feature.is_equivalent(other.top_boundary_feature):
                return False
        elif other.top_boundary_feature is not None:
            return False
        if self.base_boundary_feature is not None:
            if not self.base_boundary_feature.is_equivalent(other.base_boundary_feature):
                return False
        elif other.base_boundary_feature is not None:
            return False
        if check_extra_metadata and not equivalent_extra_metadata(self, other):
            return False
        return True

    def _load_from_xml(self):

        self.phase = rqet.find_tag_text(self.root, 'Phase')

        feature_ref_node = rqet.find_tag(self.root, 'FluidBoundaryTop')
        assert feature_ref_node is not None
        feature_root = self.model.referenced_node(feature_ref_node)
        feature_uuid = rqet.uuid_for_part_root(feature_root)
        assert feature_uuid is not None, 'rock fluid top boundary feature missing from model'
        self.top_boundary_feature = BoundaryFeature(self.model, uuid = feature_uuid)

        feature_ref_node = rqet.find_tag(self.root, 'FluidBoundaryBottom')
        assert feature_ref_node is not None
        feature_root = self.model.referenced_node(feature_ref_node)
        feature_uuid = rqet.uuid_for_part_root(feature_root)
        assert feature_uuid is not None, 'rock fluid bottom boundary feature missing from model'
        self.base_boundary_feature = BoundaryFeature(self.model, uuid = feature_uuid)

    def create_xml(self, add_as_part = True, add_relationships = True, originator = None, reuse = True):
        """Creates a rock fluid unit feature organisational xml node from this rock fluid unit feature object."""

        assert self.feature_name and self.phase and self.top_boundary_feature and self.base_boundary_feature
        if self.phase not in self.valid_phases:
            raise ValueError(f"Phase '{self.phase}' not recognized")

        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object
        # create node with citation block
        rfuf = super().create_xml(add_as_part = False, originator = originator)

        phase_node = rqet.SubElement(rfuf, ns['resqml2'] + 'Phase')
        phase_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Phase')
        phase_node.text = self.phase

        top_boundary_root = self.top_boundary_feature.root
        assert top_boundary_root is not None
        self.model.create_ref_node('FluidBoundaryTop',
                                   self.model.title_for_root(top_boundary_root),
                                   top_boundary_root.attrib['uuid'],
                                   content_type = 'obj_BoundaryFeature',
                                   root = rfuf)

        base_boundary_root = self.base_boundary_feature.root
        assert base_boundary_root is not None
        self.model.create_ref_node('FluidBoundaryBottom',
                                   self.model.title_for_root(base_boundary_root),
                                   base_boundary_root.attrib['uuid'],
                                   content_type = 'obj_BoundaryFeature',
                                   root = rfuf)

        if add_as_part:
            self.model.add_part('obj_RockFluidUnitFeature', self.uuid, rfuf)
            if add_relationships:
                self.model.create_reciprocal_relationship(rfuf, 'destinationObject', top_boundary_root, 'sourceObject')
                self.model.create_reciprocal_relationship(rfuf, 'destinationObject', base_boundary_root, 'sourceObject')

        return rfuf
